# Windfreaktech SynthHD Microwave Source
# Andre Schneider <erdna.na@gmail.com>, 2015
# Lukas Gruenhaupt <lukas.gruenhaupt@kit.edu> 05/2018
# Martin Spiecker <martin.spiecker@kit.edu> 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import logging
from time import sleep
import numpy as np
from scipy.interpolate import RectBivariateSpline
from scipy import optimize
from distutils.version import LooseVersion

path_hardcode = r'C:\qkit\qkit\drivers\Windfreak_calibration_files'  # TODO


class Windfreaktech916(Instrument):
    '''
    This is the python QTLab driver for the Windfreaktech SynthHD microwave source
    '''

    def __init__(self, name, address, model='Windfreaktech', serial='290'):
        '''
        Initializes the Windfreaktech_SynthHD, and communicates with the wrapper.
        
        Input:
            name (string)    : name of the instrument
            address (string) : address
            serial (string)  : serial number of the Windfreak source. used to identify the correct calibration file
        '''
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._model = model
        self._visainstrument = visa.instrument(self._address)
        self._serial = serial
        
        if LooseVersion(visa.__version__) < LooseVersion("1.5.0"):
            self._visainstrument.term_chars = '\n'
        else:
            self._visainstrument.write_termination = '\n'
            self._visainstrument.read_termination = '\n'
            
        self._numchannels = 2
        self._frequency = np.zeros(2)
        self._power = np.zeros(2)
        self._power_level = np.zeros(2)
        

        # Implement parameters
        self.add_parameter('frequency', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchannels), minval=54e6, maxval=13.6e9, units='Hz', channel_prefix='ch%d_')

        self.add_parameter('power_level', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchannels), minval=0, maxval=45000, units='arbu', channel_prefix='ch%d_')
    
        self.add_parameter('power', type=float,
            flags=Instrument.FLAG_GETSET,
            channels=(1, self._numchannels), minval=-70, maxval=20, units='dBm', channel_prefix='ch%d_')
        
        self.add_parameter('status', type=bool,
            flags=Instrument.FLAG_GETSET,
            channels=(1, self._numchannels), channel_prefix='ch%d_')
        
        self.add_function('get_all')
        self.load_calibration()
        
        # sets the reference (clock) frequency ###
        #self._visainstrument.write('C0x0')                 #added these two command lines to set the reference
        #self._visainstrument.write('C1x0')		           #to external for both channels
        #self.set_PLL_frequency(10e6)
        
        # switch off self-calibrated temperature compensation 
        # when tested in May 2018, this lead to jumps to the max power value some time after the power was set to a lower value
        self._visainstrument.write('C0Z0')
        self._visainstrument.write('C1Z0')

        # Unmute channels
        self._visainstrument.write('C0h1C1h1')
        self._visainstrument.write('C0h1C1h1')

        # initialize parameters
        self.get_all()



##############################       
### initialization related ###
##############################
        
    def get_all(self):

        self.get_ch1_frequency()
        self.get_ch2_frequency()
        self.get_ch1_power()
        self.get_ch2_power()
        self.get_ch1_power_level()
        self.get_ch2_power_level()
        self.get_ch1_status()
        self.get_ch2_status()
    
###########################################
### Communication with device and setup ###
###########################################
        
    def load_calibration(self):
        '''
        reloads power calibration from file Windfreak_SerialNo_Channel(A/B).cal in instruments\Windfreak_calibration_files folder.
        Calibration file was measured in frequency interval 1 GHz - 13 GHz (frequency interval 10 MHz) and with power_level_parameter 10e3 to 45e3 (steps of 500)
        '''
        
        try:

            data_ChA = np.load(
                path_hardcode + "\Windfreak_" + self._serial + "_ChannelA_calibration_with_freq_and_pow.npy",
                allow_pickle=True)
            data_ChB = np.load(
                path_hardcode + "\Windfreak_" + self._serial + "_ChannelA_calibration_with_freq_and_pow.npy",
                allow_pickle=True)

            f_channelA = data_ChA.item().get("frequency_Hz")
            f_channelB = data_ChB.item().get("frequency_Hz")

            p_channelA = data_ChA.item().get("power_intrinsic_units")
            p_channelB = data_ChB.item().get("power_intrinsic_units")

            cal_matrix_A = data_ChA.item().get("calibration_matrix_dBm")
            cal_matrix_B = data_ChB.item().get("calibration_matrix_dBm")

            # generate the interpolators to estimate the power_level parameter in set_power
            self._interp_amplitudeA = RectBivariateSpline(f_channelA, p_channelA, cal_matrix_A)
            self._interp_amplitudeB = RectBivariateSpline(f_channelB, p_channelB, cal_matrix_B)
                
        except IOError:
            raise IOError('Calibration file of Windfreak source not found')

    def set_PLL_frequency(self, frequency):
        '''
        set the reference frequency for the PLL in Hz
        and set 'x' to 0 (external reference clock)
        if frequency == False: Set clock to internal reference
        '''
        if not frequency:
            self._visainstrument.write("x1".format(27))  # unstable
        else:
            self._visainstrument.write("*{}x0".format(int(frequency/1e6)))

    def get_PLL_frequency(self):
        
        return 1e6 * float(self._visainstrument.query("*?"))
        
#######################
### functionalities ###    
#######################
    
    def do_get_frequency(self, channel):
        
        self._frequency[channel - 1] = 1e6 * float(self._visainstrument.query('C{}f?'.format(channel - 1)))
        return self._frequency[channel - 1]
        
    def do_set_frequency(self, frequency, channel):
        '''
        Set frequency of device

        Input:
            freq (float) : Frequency in Hz

        Output:
            None
        '''
        self._frequency[channel - 1] = frequency
        self._visainstrument.write('C{}f{:.6f}'.format(channel - 1, frequency / 1e6))
        # After every frequency change the power level has to be resetted
        # Since the output power is frequency dependent, the calibrated power is chosen
        # self.do_set_power(self._power[channel - 1], channel)
        
        # sleep since source needs some time to adjust to new settings
        sleep(0.05)

    def do_set_frequency_uncalibrated(self, frequency, channel):
        '''
        Set frequency of device

        Input:
            freq (float) : Frequency in Hz

        Output:
            None
        '''
        self._frequency[channel - 1] = frequency
        self._visainstrument.write('C{}f{:.6f}'.format(channel - 1, frequency / 1e6))
        self._visainstrument.write('C{}a{}'.format(channel - 1, self._power_level[channel - 1]))

        # sleep since source needs some time to adjust to new settings
        sleep(0.05)
    
    def do_get_power_level(self, channel):

        self._power_level[channel - 1] = self._visainstrument.query('C{}a?'.format(channel - 1))
        return self._power_level[channel - 1]
        
    def do_set_power_level(self, level, channel):
        '''
        Set uncalibrated power level of device
        Usually, you should use set_power which accepts values in dBm

        Input:
            level (float) : (0=mimimum, 45000=maximum)

        Output:
            None
        '''
        self._power_level[channel - 1] = level
        self._visainstrument.write('C{}a{}'.format(channel - 1, level))
        
        # sleep since source needs some time to adjust to new settings
        sleep(0.05)
    
    # TODO load files
    def do_get_power(self, channel):
        '''
        this only returns the stored power value and does not communicate with the device!
        '''
        
        if self._power[channel - 1] is None:
            
            print("Power is not calibrated. You have to do this to get a proper value.")
            
        return self._power[channel - 1]
        
    def do_set_power(self, power, channel):
        '''
        Set output power of device

        Input:
            power (float) : Power in dBm

        Output:
            None
        '''
        
        self._power[channel - 1] = power
        
        if channel == 0:
            
            # generate a helper function to find power level for desired power in dBm by finding roots of the function
            helper_fn = lambda x: self._interp_amplitudeA(self._frequency[channel - 1], x) - power
            # by finding the zero of the helper function, the power level value is determined
            pwr_lvl = int(optimize.newton(helper_fn, 20000))
            
            if pwr_lvl > 45000:
                
                print('The minimum and maximum output power of this microwave source depends on frequency. '
                      'Your desired combination of frequency and output power is not possible. '
                      'Power is set to max for this frequency: {}.1f dBm'.format(self._interp_amplitudeA(self._frequency[channel - 1], 45000)))
            else:
                
                self.do_set_power_level(pwr_lvl, channel)
                
        if channel == 1:
            
            # generate a helper function to find power level for desired power in dBm by finding roots of the function
            helper_fn = lambda x: self._interp_amplitudeB(self._frequency[channel - 1], x) - power
            # by finding the zero of the helper function, the power level value is determined
            pwr_lvl = int(optimize.newton(helper_fn, 20000))
            
            if pwr_lvl > 45000:
                
                print ('The minimum and maximum output power of this microwave source depends on frequency. '
                       'Your desired combination of frequency and output power is not possible. '
                       'Power is set to max for this frequency: {}.1f dBm'.format(self._interp_amplitudeB(self._frequency[channel - 1], 45000)))
            else:
                
                self.do_set_power_level(pwr_lvl, channel)

    def do_get_status(self, channel):

        return bool(int(float(self._visainstrument.query('C{}r?'.format(channel - 1)))))
        
    def do_set_status(self, status, channel):
        '''
        Set status of output channel
        Sets 'r' (output-parameter)  as well as 'E' (PLL Chip Enable) according to the manual

        Input:
            channel (int) : 0 for channel A, 1 for channel B
            status        : True or False 

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting status to "{}s"'.format(status))
        
        if status:
             
            self._visainstrument.write('C{}r1E1'.format(channel - 1))
            
        else:
            
            self._visainstrument.write('C{}r0E0'.format(channel - 1))

    def write(self, msg):
        return self._visainstrument.write(msg)    

    def query(self, msg):
        return self._visainstrument.query(msg)



# AnaPico_APUASYN20.py
# Nicolas Gosling <Nicolas.Gosling@partner.kit.edu> 04/21
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
import sys
from qkit import visa

import logging
import numpy

class AnaPico_APUASYN20(Instrument):
    '''
    This is the driver for the Ana Pico APUASYN20 Signal Genarator

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'AnaPico_APUASYN20', address='<GBIP address>, reset=<bool>')
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the Ana Pico APUASYN20, and communicates with the wrapper.

        Input:
          name (string)    : name of the instrument
          address (string) : GPIB address
          reset (bool)     : resets to default values, default=False
        '''
        logging.info(__name__ + ' : Initializing instrument Ana Pico APUASYN20')
        super().__init__(name, tags=['physical'])


        # Add some global constants
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        self._numchannels = 4

        self.add_parameter('power',
            flags=super().FLAG_GETSET, units='dBm', minval=-10, maxval=23,
            channels=(1, self._numchannels), channel_prefix='ch%d_', type=float)
        self.add_parameter('frequency',
            flags=super().FLAG_GETSET, units='Hz', minval=1e5, maxval=20e9,
            channels=(1, self._numchannels), channel_prefix='ch%d_', type=float)
        self.add_parameter('status', flags=super().FLAG_GETSET,
            channels=(1, self._numchannels), channel_prefix='ch%d_', type=bool)
        self.add_parameter('blanking', flags=super().FLAG_GETSET,
            channels=(1, self._numchannels), channel_prefix='ch%d_', type=bool)
        #self.add_parameter('channel',
            #flags=super().FLAG_GETSET, type=int, minval = 1, maxval = 4)



    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : get all')
        self.get_ch1_frequency()
        self.get_ch2_frequency()
        self.get_ch3_frequency()
        self.get_ch4_frequency()
        self.get_ch1_power()
        self.get_ch2_power()
        self.get_ch3_power()
        self.get_ch4_power()
        self.get_ch1_status()
        self.get_ch2_status()
        self.get_ch3_status()
        self.get_ch4_status()
        self.get_ch1_blanking()
        self.get_ch2_blanking()
        self.get_ch3_blanking()
        self.get_ch4_blanking()


    def do_get_status(self, channel):
        '''
        Reads the output status from the instrument

        Input:
            None

        Output:
            status (string) : 'On' or 'Off'
        '''
        logging.debug(__name__ + f' : get status of channel {channel}')
        return self._visainstrument.query(f'OUTPut{channel}?')


    def do_set_status(self,  status, channel):
        '''
        Set the output status of the instrument

        Input:
            status (bool) : True or False

        Output:
            None
        '''
        logging.debug(__name__ + f' : set status of channel {channel} to {status}')

        if status == True:
            self._visainstrument.write(f'OUTPut{channel} ON')
        elif status == False:
            self._visainstrument.write(f':OUTPut{channel} OFF')
        else:
            raise ValueError('set_status(): can only set True or False')

    def do_set_blanking(self, channel, blanking_status=False):
        '''
        Set the output blanking of the instrument (blanking means the ouput will be turned off when the frequency changes)

        Input:
            status (bool) : True or False

        Output:
            None
        '''
        logging.debug(__name__ + f' : set blanking to {blanking_status}')

        if blanking_status == True:
            self._visainstrument.write(f'OUTPut{channel}:BLANking ON')
        elif blanking_status == False:
            self._visainstrument.write(f':OUTPut{channel}:BLANking OFF')
        else:
            raise ValueError('set_status(): can only set True or False')


    def do_get_blanking(self, channel):
        '''
        Get the output blanking of the instrument (blanking means the output will be turned off when the frequency changes)

        Input:
            NONE

        Output:
            bool True, False
        '''
        logging.debug(__name__ + f' : get blanking status of channel: {channel}')

        return self._visainstrument.query(f':OUTPut{channel}:BLANking?')



    def do_get_channel(self):
        '''
        Reads the channel selection from the instrument

        Input:
            None

        Output:
            default channel number
        '''

        logging.debug(__name__ + ' : get channel')
        return self._visainstrument.query('SELect?')


    def do_set_channel(self, channel):
        '''
        Sets the channel selection from the instrument

        Input:
            channel (int): 1,2,3, or 4

        Output:
            NONE
        '''
        self._channel = channel
        logging.debug(__name__ + f' : set channel to {channel}')
        self._visainstrument.write(f'SELect {channel}')


    def do_get_frequency(self, channel):
        '''
        Reads the frequency of the signal from the instrument

        Input:
            None

        Output:
            freq (float) : Frequency in Hz
        '''
        logging.debug(__name__ + f' : get frequency of channel {channel}')
        return self._visainstrument.query(f':SOURce{channel}:FREQuency?')

    def do_set_frequency(self, freq, channel):
        '''
        Set the frequency of the instrument

        Input:
            freq (float) : Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + f' : set frequency of channel {channel} to {freq}')
        self._visainstrument.write(f':SOURce{channel}:FREQuency {freq}')


    def do_get_power(self, channel):
        '''
        Reads the power of the signal from the instrument

        Input:
            None

        Output:
            ampl (?) : power in ?
        '''
        logging.debug(__name__ + f' : get power of channel {channel}')
        return float(self._visainstrument.query(f':SOURce{channel}:POWer?'))


    def do_set_power(self, amp, channel):
        '''
        Set the power of the signal

        Input:
            amp (float) : power in ??

        Output:
            None
        '''
        logging.debug(__name__ + f' : set power of channel {channel} to {amp}')
        self._visainstrument.write(f':SOURce{channel}:POWer {amp}')


# Oxford_Triton.py
# MP@KIT 09/2016
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
import socket
import logging


class Oxford_Triton(Instrument):
    '''
    This is the python driver for the Oxford Triton 200 fridge

    Usage:
    Initialise with
    <name> = instruments.create('<name>', host='<IP address>', port=<port>)
    For the bypass functionality, you need to specify the RaspberryPi as HOST,
    where a proxy server runs, passing all commands to the Triton control PC.

    '''
    
    def __init__(self, name, host, port):
        '''
        Initializes

        Input:
            name (string)    : name of the instrument
            address (string) : IP address
            port (int)       : port
        '''
        
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])
        
        self._host = host
        self._port = port
        self._soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._soc.connect((self._host, self._port))
        
        # Implement parameters
        self.add_parameter('temperature', type=float,
                           flags=Instrument.FLAG_GET,
                           minval=0, maxval=350,
                           units='K')
        
        self.add_parameter('resistance', type=float,
                           flags=Instrument.FLAG_GET,
                           minval=0, maxval=200000,
                           units='Ohm')
        
        self.add_parameter('pressure', type=float,
                           flags=Instrument.FLAG_GET,
                           minval=0, maxval=5,
                           units='mbar')
        
        self.add_parameter('valve', type=bool,
                           flags=Instrument.FLAG_GETSET)
        
        self.add_parameter('pulse_tube', type=bool,
                           flags=Instrument.FLAG_GETSET)
        
        self.add_parameter('still_power', type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=0, maxval=1,
                           units='uW')
        
        self.add_parameter('base_power', type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=0, maxval=300,
                           units='uW')
        
        self.add_parameter('warm_up_heater', type=bool,
                           flags=Instrument.FLAG_GETSET)
        
        self.add_parameter('pump', type=bool, flags=Instrument.FLAG_GETSET)
        self.add_parameter('bypass', type=bool, flags=Instrument.FLAG_GETSET)
        self.add_parameter('turbo_speed', type=int, flags=Instrument.FLAG_GET)
        self.add_parameter('turbo_power', type=int, flags=Instrument.FLAG_GET)
        
        # Implement functions
        self.add_function('get_all')
        self.add_function('start_pre_cooling')
        self.add_function('empty_pre_cooling')
        self.add_function('start_condensing')
        self.add_function('cool_down')
        self.add_function('warm_up')
        self.add_function('get_status')
        self.add_function('get_automatisation')
        self.add_function('stop_automatisation')
        self.add_function('get_base_control')
        self.add_function('set_base_control')
        
        self.get_all()
    
    def get_all(self):
        for i in range(7):
            self.get_temperature(channel=i + 1)
        for i in range(5):
            self.get_pressure(gauge=i + 1)
        for i in range(8):
            self.get_valve(valve=i + 1)
        self.get_still_power()
        self.get_base_power()
        self.get_warm_up_heater()
    
    ###
    # Communication with device
    ###
    
    def start_pre_cooling(self):
        '''
        Starts pre cooling

        Input:
            None

        Output:
            bool as confirmation
        '''
        logging.debug(__name__ + ' : start pre-cooling')
        ret = self._ask('SET:SYS:DR:ACTN:PCL')
        return ret
    
    def empty_pre_cooling(self):
        '''
        Emptys pre cooling

        Input:
            None

        Output:
            bool as confirmation
        '''
        logging.debug(__name__ + ' : empty pre-cool circuit')
        ret = self._ask('SET:SYS:DR:ACTN:EPCL')
        return ret
    
    def start_condensing(self):
        '''
        Starts condensing

        Input:
            None

        Output:
            bool as confirmation
        '''
        logging.debug(__name__ + ' : start condensing')
        ret = self._ask('SET:SYS:DR:ACTN:COND')
        return ret
    
    def cool_down(self):
        '''
        Starts full cool down

        Input:
            None

        Output:
            bool as confirmation
        '''
        logging.debug(__name__ + ' : cool down')
        ret = self._ask('SET:SYS:DR:ACTN:CLDN')
        return ret
    
    def warm_up(self):
        '''
        Starts warm up

        Input:
            None

        Output:
            bool as confirmation
        '''
        logging.debug(__name__ + ' : warm up')
        # self.set_warm_up_heater(True)
        ret = self._ask('SET:SYS:DR:ACTN:WARM')
        return ret
    
    def get_status(self):
        '''
        Get status

        Input:
            None

        Output:
            status (str)
        '''
        logging.debug(__name__ + ' : get status')
        return self._ask('READ:SYS:DR:STATUS').strip()[19:]
    
    def get_automatisation(self):
        '''
        Get name of running automatisation

        Input:
            None

        Output:
            automatisation (str)
        '''
        logging.debug(__name__ + ' : get name of running automatisation')
        return self._ask('READ:SYS:DR:ACTN').strip()[17:]
    
    def stop_automatisation(self):
        '''
        Stop automatisation

        Input:
            None

        Output:
            bool as confirmation
        '''
        logging.debug(__name__ + ' : stop automatisation')
        return self._ask('SET:SYS:DR:ACTN:STOP').strip()[26:] == 'VALID'
    
    def get_base_control(self):
        '''
        Get base temperature

        Input:
            None
        Output:
            status (bool)
            P (float)
            I (int)
            D (int)
            temperature (float)
        '''
        if self._ask('READ:DEV:T5:TEMP:LOOP:MODE').strip()[27:] == 'NOT_FOUND':
            status = self._ask('READ:DEV:T8:TEMP:LOOP:MODE').strip()[27:] == 'ON'
            t_set = float(self._ask('READ:DEV:T8:TEMP:LOOP:TSET').strip()[27:-1])
            P = float(self._ask('READ:DEV:T8:TEMP:LOOP:P').strip()[24:])
            I = float(self._ask('READ:DEV:T8:TEMP:LOOP:I').strip()[24:])
            D = float(self._ask('READ:DEV:T8:TEMP:LOOP:D').strip()[24:])
        if self._ask('READ:DEV:T8:TEMP:LOOP:MODE').strip()[27:] == 'NOT_FOUND':
            status = self._ask('READ:DEV:T5:TEMP:LOOP:MODE').strip()[27:] == 'ON'
            t_set = float(self._ask('READ:DEV:T5:TEMP:LOOP:TSET').strip()[27:-1])
            P = float(self._ask('READ:DEV:T5:TEMP:LOOP:P').strip()[24:])
            I = float(self._ask('READ:DEV:T5:TEMP:LOOP:I').strip()[24:])
            D = float(self._ask('READ:DEV:T5:TEMP:LOOP:D').strip()[24:])
        else:
            return 'False'
        return status, t_set, P, I, D
        logging.debug(__name__ + ' : getting base controle ')
        return float(self._ask(''))
    
    def set_base_control(self, temperature=0, P=12., I=1, D=0, status=True):
        '''
        Set base controle to temperature usint PID values

        Input:
            temperature (float), optional, default = 0
            P (float), optional, default = 12.
            I (int), optional, default = 100.
            D (int), optional, default = 0.
            status (bool), optional, default = True
        Output:
            bool as confirmation
        '''
        
        if status:
            thermometer = 8
            logging.debug(__name__ + ' : setting base temperature to %f using P %f I %i D%i' % (temperature, P, I, D))
            if temperature >= 2.2:
                thermometer = 5
                self.set_pulse_tube(False)
                self.set_still_power(0)
                self.set_valve(9, True)
                self.set_valve(4, True)
            return str(self._ask('SET:DEV:T%i:TEMP:LOOP:HTR:H1:MODE:ON:P:%f:I:%i:D:%i:TSET:%f' % (thermometer, P, I, D, temperature))).find('INVALID') == -1
        else:
            logging.debug(__name__ + ' : turn off base control')
            if self._ask('READ:DEV:T5:TEMP:LOOP:MODE').strip()[27:] == 'ON':
                return self._ask('SET:DEV:T5:TEMP:LOOP:MODE:OFF').strip()[35:] == 'VALID'
            if self._ask('READ:DEV:T8:TEMP:LOOP:MODE').strip()[27:] == 'ON':
                return self._ask('SET:DEV:T8:TEMP:LOOP:MODE:OFF').strip()[35:] == 'VALID'
            else:
                return 'False'
    
    ###
    # GET and SET functions
    ###
    
    
    def do_get_temperature(self, channel=8):
        '''
        Get temperature of thermometer at channel

        Input:
            channel (int), default = 8, RuOx at MC
        Output:
            temperature (float)
        '''
        logging.debug(__name__ + ' : getting temperature of channel %i' % channel)
        return float(self._ask('READ:DEV:T%i:TEMP:SIG:TEMP' % channel).strip()[26:-1])
    
    def do_get_resistance(self, channel=8):
        '''
        Get resistance of thermometer at channel

        Input:
            channel (int), default = 8, RuOx at MC
        Output:
            resistance (float)
        '''
        logging.debug(__name__ + ' : getting resistance of channel %i' % channel)
        return float(self._ask('READ:DEV:T%i:TEMP:SIG:RES' % channel).strip()[25:-3])
    
    def do_get_pressure(self, gauge=1):
        '''
        Get pressure at gauge

        Input:
            pressure_meter (int), default=1, tank
        Output:
            pressure (float)
        '''
        
        logging.debug(__name__ + ' : getting pressure at gauge %i' % gauge)
        return float(self._ask('READ:DEV:P%i:PRES:SIG:PRES' % gauge).strip()[26:-2])
    
    def do_get_valve(self, valve):
        '''
        Get valve status at valve

        Input:
            valve number (int)
        Output:
            status (bool)
        '''
        status_dict = {'CLOSE': 0, 'OPEN': 1, 'TOGGLE': 2}
        logging.debug(__name__ + ' : getting valve %s status' % valve)
        return status_dict[self._ask('READ:DEV:V%i:VALV:SIG:STATE' % valve).strip()[27:]]
    
    def do_set_valve(self, valve, status):
        '''
        Set valve status at valve_number to status

        Input:
            valve number (int)
            status (bool)
        Output:
            bool as confirmation
        '''
        status_dict = {0: 'CLOSE', 1: 'OPEN', 2: 'TOGGLE'}
        return self._ask('SET:DEV:V%i:VALV:SIG:STATE:%s' % (valve, status_dict[status])).strip()[27:] == status_dict[status]
    
    def do_get_pulse_tube(self):
        '''
        Get pulse tube status

        Input:
            None
        Output:
            status (bool)
        '''
        status_dict = {'OFF': 0, 'ON': 1}
        logging.debug(__name__ + ' : getting pulse tube status')
        try:
            return self._ask('READ:DEV:C1:PTC:SIG:STATE').strip()[26:]
        except Exception:
            return False
    
    def do_set_pulse_tube(self, status):
        '''
        Set pulst tube to status

        Input:
            status (bool)
        Output:
            bool as confirmation
        '''
        status_dict = {0: 'OFF', 1: 'ON'}
        logging.debug(__name__ + ' : setting pulse tube to %i' % status)
        return self._ask('SET:DEV:C1:PTC:SIG:STATE:%s' % status_dict[status]).strip()[27:] == status_dict[status]
    
    def do_get_still_power(self):
        '''
        Get still heater power

        Input:
            None
        Output:
            power (float)
        '''
        
        logging.debug(__name__ + ' : getting still power')
        return float(self._ask('READ:DEV:H2:HTR:SIG:POWR').strip()[25:-2])
    
    def do_set_still_power(self, power):
        '''
        Set still heater to power

        Input:
            power (float)
        Output:
            bool as confirmation
        '''
        
        logging.debug(__name__ + ' : setting still heater power to %f' % power)
        add = len(str(power).split('.')[0])
        return self._ask('SET:DEV:H2:HTR:SIG:POWR:%f' % power).strip()[37 + add:] == 'VALID'
    
    def do_get_base_power(self):
        '''
        Get base heater power

        Input:
            None
        Output:
            power (float)
        '''
        
        logging.debug(__name__ + ' : getting base power')
        return float(self._ask('READ:DEV:H1:HTR:SIG:POWR').strip()[25:-2])
    
    def do_set_base_power(self, power):
        '''
        Set base heater to power

        Input:
            power (float)
        Output:
            bool as confirmation
        '''
        
        logging.debug(__name__ + ' : setting base heater power to %f' % power)
        add = len(str(power).split('.')[0])
        return self._ask('SET:DEV:H1:HTR:SIG:POWR:%f' % power).strip()[37 + add:] == 'VALID'
    
    def do_get_warm_up_heater(self):
        '''
        Get status of warm up heater

        Input:
            None
        Output:
            status (bool)
        '''
        
        logging.debug(__name__ + ' : getting warm up heater status')
        return float(self._ask('READ:DEV:H3:HTR:SIG:POWR').strip()[25:-2]) != 0
    
    def do_set_warm_up_heater(self, status):
        '''
        Set warm up heater to status

        Input:
            status (bool)
        Output:
            bool as confirmation
        '''
        
        logging.debug(__name__ + ' : setting warm up heater to %i' % status)
        power = 0
        if status:
            power = 10000
        add = len(str(power).split('.')[0])
        self._ask('SET:DEV:H3:HTR:SIG:POWR:%f' % power).strip()[37 + add:]
        return self.get_warm_up_heater() is status
    
    def _ask(self, cmd):
        self._soc.sendall((cmd + '\n').encode())
        return self._soc.recv(1024).decode()
    
    def _do_set_pump(self, pump="COMP", state=False):
        '''
        Set pump on or off

        Input:
            pump: one of 'TURB1', 'FP', 'COMP'
            state (bool)
        '''
        logging.debug(__name__ + ' : setting pump {} to state {}'.format(pump, "ON" if state else "OFF"))
        return self._ask('SET:DEV:{}:PUMP:SIG:STATE:{}'.format(pump, "ON" if state else "OFF")).strip()[-6:] == ":VALID"
    
    def _do_get_pump(self, pump="COMP"):
        '''
        Get pump state (bool)

        Input:
            pump: one of 'TURB1', 'FP', 'COMP'
        Output:
            status (bool)
        '''
        logging.debug(__name__ + ' : Getting pump {} state'.format(pump))
        return self._ask('READ:DEV:{}:PUMP:SIG:STATE'.format(pump)).strip()[-2:] == "ON"
    
    def _do_set_bypass(self, state):
        '''
        Open or close bypass and switch compressor accordingly.
        Does a safety check wether P2<750mbar

        Input:
            state (bool): True: open bypass and switch off compressor,
                          False: close bypass and switch on compressor
        Output:
            None
        '''
        logging.debug(__name__ + ' : {} bypass'.format("open" if state else "close"))
        if state:  # opening, do safety check
            if self.do_get_pressure(gauge=2) > 750:
                raise ValueError('{} bypass not successful. P2 is above 750mbar.'.format("open" if state else "close"))
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self._host, 9989))
            sock.send(b"open\n")
            response = sock.recv(1024 * 8).strip().decode()
            if response == "Ok":
                self._do_set_pump("COMP", False)
            else:
                raise ValueError("Opening bypass not successful '{}' Compressor still on".format(response))
        else:  # closing
            self._do_set_pump("COMP", True)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self._host, 9989))
            sock.send(b"close\n")
            response = sock.recv(1024 * 8).strip().decode()
            if response == "Ok":
                return True
            else:
                return False
    
    def _do_get_bypass(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self._host, 9989))
        sock.send(b"get_status\n")
        response = sock.recv(1024 * 8).strip().decode()
        if response == "open":
            return True
        if response == "closed":
            return False
        else:
            raise valueError("get_bypass responded with '{}'".format(response))
    
    def _do_get_turbo_speed(self):
        logging.debug(__name__ + ' : Getting speed of the turbo pump')
        return int(self._ask('READ:DEV:TURB1:PUMP:SIG:SPD').strip()[28:-2])
    
    def _do_get_turbo_power(self):
        logging.debug(__name__ + ' : Getting power of the turbo pump')
        return int(self._ask('READ:DEV:TURB1:PUMP:SIG:POWR').strip()[29:-1])


import qkit
from qkit.core.instrument_base import Instrument
import logging
import serial
import time, numpy as np
import threading


class Faulhaber_3564K024B_C(Instrument):
    """
    This is the driver for the Faulhaber 3564K024B C stepping motor.
    """

    def __init__(self, name, address):
        """
        Initializes serial communication with the instrument Faulhaber 3564K024B C.

        Parameters
        ----------
        name: string
            Name of the instrument (driver).
        address: string
            Serial address for the communication with the instrument.

        Returns
        -------
        None
        """
        Instrument.__init__(self, name, tags=['physical'])
        logging.info(__name__ + ': Initializing instrument Faulhaber 3564K024B C')
        # Serial port configuration
        baudrate = 9600
        timeout = 0.1
        self.ins = serial.Serial(address, baudrate, timeout=timeout)
        self.running, self.aborting = False, False
        self.lock = threading.Lock()
        
        self.mode = 'SOR0'
        self.set_mode(self.mode)
        self.position = None
        self.speed = 10000
        self.direction = 0
        self.bounds = (-1, 1500000)
        self.set_bounds(bounds=self.bounds)

    def _write(self, cmd):
        """
        Sends a serial command <cmd> to the Device.

        Parameters
        ----------
        cmd: str
            Command that is send to the instrument.

        Returns
        -------
        None
        """
        with self.lock:
            return self.ins.write('{:s}\r'.format(cmd).encode())

    def _query(self, cmd, size=100):
        """
        Sends a serial command <cmd> and returns the read answer <ans>.

        Parameters
        ----------
        cmd: str
            Command that is send to the instrument.

        Returns
        -------
        answer: str
            Answer that is returned at read after the sent <cmd>.
        """
        with self.lock:
            self.ins.write('{:s}\r'.format(cmd).encode())
            return self.ins.read(size=size).decode().split('\r\n')[0] # strip('\r\n') # 
    
    def set_mode(self, mode):
        """
        Sets the motor operation mode <mode>.

        Parameters
        ----------
        mode: str
            Motor operation mode. Can be:
            'SOR': source for velocity by the 
                0: interface (default)
                1: voltage at the analog input
                2: pwm signal at the analog input
                3: current setpoint at the analog input
                4: current setpoint at the analog input with the direction by polarity
            'CONTMOD': continuous mode
            'STEPMOD': step motor mode
            'APCMOD': analog position control mode
            'ENCMOD': encoder mode
            'HALLSPEED': hall sensor as speed sensor
            'ENCSPEED': encoder as speed sensor
            'GEARMOD': gearing mode
            'VOLTMOD': set voltage mode
            'IXRMOD': set IxR mode

        Returns
        -------
        None
        """
        self.mode = mode
        self._write(mode)
        
    def set_speed(self, speed, direction=None):
        """
        Sets motor rotation speed to <speed> and the operating mode to <SOR0>, which is needed.

        Parameters
        ----------
        speed: int
            Motor roation speed (1rpm equals about 1500). Must be in [0, 30 000].

        Returns
        -------
        None
        """
        try:
            self.speed = speed
            if direction != None:
                self.direction = direction
            self.set_direction(val=self.direction)
            self._write('SOR0')
            self._write('V{:s}{:d}'.format({0: '+', 1: '-'}[direction], self.speed))
        except:
            self.set_speed(speed=speed, direction=direction)

    def set_status(self, status):
        """
        Sets motor status to <status>.

        Parameters
        ----------
        status: int
            Motor status. Possible values are 0 (off) and 1 (on).

        Returns
        -------
        None
        """
        try:
            if self.mode == 'SOR0':
                self.set_speed(speed={True: self.speed, False: 0}[status], direction=self.direction)
            self._write({True: 'EN', False: 'DI'}[status])
        except:
            self.set_status(status=status)

    def set_direction(self, val):
        """
        Sets motor rotation direction to <val>.

        Parameters
        ----------
        val: int
            Motor rotation direction. Possible values are 0 (right) and 1 (left).

        Returns
        -------
        None
        """
        self.direction = val
        try:
            self._write({0: 'ADR', 1: 'ADL'}[val])
        except:
            self.set_direction(val=val)

    def get_position(self):
        """
        Gets the motor position <ans>.

        Parameters
        ----------
        None

        Returns
        -------
        ans: int
            Motor position.
        """
        try:
            self.position = float(self._query('pos'))
        except ValueError:
            print('''Faulhaber_3564K024B_C: couldn't get position.''')
        finally:
            return self.position

    def set_zero_position(self):
        """
        Sets the motor's actual position to 0.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self._write('HO')

    def move_to(self, pos, t=1e-2):
        """
        Moves the motor to the positions <pos>.

        Parameters
        ----------
        pos: int
            Motor position.
        t: float
            Query interval for actual position.

        Returns
        -------
        pos: int
            Actual position
        """
        if self.mode == 'SOR1':
            def _move_to():
                rel = int(np.sign(self.get_position() - pos))
                self.set_direction({1: 1, -1: 0}[rel])
                self.set_status(True)
                while {1: self.get_position() > pos, -1: self.get_position() < pos}[rel] and not self.aborting:
                    time.sleep(t)
                self.set_status(False)
        elif self.mode == 'SOR0':
            print('''use 'SOR1' mode instead''')
            raise NotImplementedError(''''move_to' not yet implemented for 'SOR0' mode.''' )
        
        self.aborting = False
        self.thread_move_to = threading.Thread(target=_move_to)
        self.thread_move_to.start()

    def set_bounds(self, bounds=(0, 1510000)):
        """
        Sets the position range to <bounds>.

        Parameters
        ----------
        bounds: tuple(int)
            Motor position range.

        Returns
        -------
        None
        """
        self._write('LL{:+d}'.format(bounds[0]))
        self._write('LL{:+d}'.format(bounds[1]))
        self._write('APL1')
    
    def get_bounds(self):
        """
        Gets the position range <bounds>.

        Parameters
        ----------
        None

        Returns
        -------
        bounds: tuple(int)
            Motor position range.
        """
        return (float(self._query('GNL')), float(self._query('GPL')))

    def run(self, bounds=None, speed=None, t=1e-2):
        """
        Runs alternating one turn to the right and to the left bound in an endless loop or until it is stopped by 'self.stop()'

        Parameters
        ----------
        bounds: tuple
            left and right bound

        Returns
        -------
        None
        """
        if bounds == None:
            bounds = self.bounds
        if speed == None:
            speed = self.speed
        
        if self.mode == 'SOR1':
            def run_endless():
                while self.running:
                    self.set_direction(0)
                    time.sleep(0.5)
                    self.set_status(True)
                    while self.get_position() < bounds[1] and not self.aborting:
                        time.sleep(t)
                    self.set_status(False)
                    if self.aborting: break
                    self.set_direction(1)
                    time.sleep(0.5)
                    self.set_status(True)
                    while self.get_position() > bounds[0] and not self.aborting:
                        time.sleep(t)
                    self.set_status(False)
                    self.set_direction(1)
        elif self.mode == 'SOR0':
            def run_endless():
                while True:  # ugly bugfix, but I don't know, how to solve the problem of a restart (MMW)
                    while self.running:
                        self.set_status(True)
                        self.set_speed(speed=speed, direction=0)
                        while self.get_position() < bounds[1] and not self.aborting:
                            time.sleep(t)
                        self.set_speed(speed=speed, direction=1)
                        while self.get_position() > bounds[0] and not self.aborting:
                            time.sleep(t)
                        self.set_speed(speed=0, direction=1)
                        self.set_status(False)
                    time.sleep(1)
        
        self.running, self.aborting = True, False
        self.thread_run = threading.Thread(target=run_endless)
        self.thread_run.start()

    def stop(self):
        """
        Stops the endless loop initiated by 'self.run()' after the next cycle, when the motor is at its start position.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self.running = False
    
    def abort(self):
        """
        Stops 'self.move_to()' or the endless loop initiated by 'self.run()' immediately.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self.running, self.aborting = False, True
        time.sleep(0.5)
        with self.lock:
            self.ins.read(size=100)
        return self.get_position()
    
    def get_running_status(self):
        """
        Gets the status of endless loop initiated by 'self.run()' and eventually stopped by 'self.stop()' or 'self.abort()'.

        Parameters
        ----------
        None

        Returns
        -------
        status: boolean
            Running state.
        """
        return self.running and not self.aborting

    def close(self):
        """
        Closes the serial connecttion with the instrument.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self.ins.close()


# wrapper for TwoPortSwitch_RT by JB@KIT 04/2015 jochen.braumueller@kit.edu
# two port switch at room temperature to in situ (re)calibrate the qubit manipulation IQ-mixer
# model: Raspberry Pi running Raspian, Radiall Switch R572.432.000 (latching, no 50Ohms termination at open port) integrated in rack slot
# use: connect mixer output line to 'C', switchable connectors 1,2 to the inlet of the cryostat and the spectrum analyzer

from qkit.core.instrument_base import Instrument
import types
import logging
#import numpy as np
import time,sys
import socket
import sys

class two_port_switch(Instrument):

    def __init__(self, name, host='10.22.197.88', port=9988):

        Instrument.__init__(self, name, tags=['physical'])
        
        self.add_parameter('position',
            type=str,
            flags=Instrument.FLAG_GETSET)
            
        self.HOST, self.PORT = host, port
    

    def do_set_position(self, position):
        if position == '1' or position == '2':
            try:
                # Create a socket (SOCK_STREAM means a TCP socket)
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                #connect
                sock.connect((self.HOST, self.PORT))
                sock.sendall('switch ' + str(position) + '\n')
                #receive confirmation from server and shut down connection
                rec = sock.recv(1024)
                #print str(rec)
                sock.close()
            except Exception as detail:
                print "Error: ", detail
                return 0
            
            if rec == 'Ok':   #switching successful
                print 'successfully switched to port #'+position
                return True
            else:
                return False
            
        else:
            print 'invalid given switch setting...no action'
            return False
            
    def do_get_position(self):
        try:
            # Create a socket (SOCK_STREAM means a TCP socket)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            #connect
            sock.connect((self.HOST, self.PORT))
            sock.sendall('get_position' + '\n')
            #receive data from the server and shut down connection
            rec = sock.recv(1024)
            print rec
            sock.close()
            return str(rec)
        except Exception as detail:
            print "Error: ", detail
            return 0
    
    #shortcuts
    def ch1(self):
        self.do_set_position('1')
    
    def ch2(self):
        self.do_set_position('2')
 



# VNA dummy driver, M. Wildermuth
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, query to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
import numpy as np

def get_resonance_curve(f, f_r, Q_l, Q_c, mode='notch'):
    if mode == 'notch':
        S_21 = 1 - (Q_l / Q_c) / (1 + 2j * Q_l * (f - f_r) / f_r)
    elif mode == 'reflection':
        S_21 = 1 - (2 * Q_l / Q_c) / (1 + 2j * Q_l * (f - f_r) / f_r)
    elif mode == 'inline':
        S_21 = (Q_l / Q_c) / (1 + 2j * Q_l * (f - f_r) / f_r)
    return S_21


class VNA_dummy(Instrument):
    """
    This is the python driver for a dummy Vector Network Analyzer

    Usage:
    Initialise with
    <name> = instruments.create('<name>')

    """

    def __init__(self, name):
        """
        Initializes

        Input:
        """
        self.__name__ = __name__
        Instrument.__init__(self, name, tags=['virtual'])
        self.func = get_resonance_curve
        self.args = ()
        self.kwargs = {'f_r': 5e9, 'Q_l': 9e3, 'Q_c': 1e4}

        # Implement parameters
        self.add_parameter('averages',
                           type=int,
                           flags=Instrument.FLAG_GETSET,
                           minval=1,
                           maxval=65536,
                           tags=['sweep'])

        self.add_parameter('Average',
                           type=bool,
                           flags=Instrument.FLAG_GETSET)

        self.add_parameter('startfreq',
                           type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=0,
                           maxval=20e9,
                           units='Hz',
                           tags=['sweep'])

        self.add_parameter('stopfreq',
                           type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=0,
                           maxval=20e9,
                           units='Hz',
                           tags=['sweep'])

        self.add_parameter('centerfreq',
                           type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=0,
                           maxval=20e9,
                           units='Hz',
                           tags=['sweep'])

        self.add_parameter('span',
                           type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=0,
                           maxval=20e9,
                           units='Hz',
                           tags=['sweep'])

        self.add_parameter('nop',
                           type=int,
                           flags=Instrument.FLAG_GETSET,
                           minval=1,
                           maxval=100001,
                           tags=['sweep'])

        self.add_parameter('sweep_type',
                           type=str,
                           flags=Instrument.FLAG_GETSET,
                           tags=['sweep'])

        self.add_parameter('sweeptime',
                           type=float,
                           flags=Instrument.FLAG_GET,
                           minval=0,
                           maxval=1e3,
                           units='s',
                           tags=['sweep'])
            
        self.add_parameter('sweeptime_averages',
                           type=float,
                           flags=Instrument.FLAG_GET,
                           minval=0,
                           maxval=1e3,
                           units='s',
                           tags=['sweep'])

        # Implement functions
        self.add_function('avg_clear')
        self.add_function('avg_status')
        self.add_function('get_freqpoints')
        self.add_function('get_tracedata')
        self.add_function('pre_measurement')
        self.add_function('start_measurement')
        self.add_function('post_measurement')
        self.add_function('ready')

        self.centerfreq = self.kwargs.get('f_r')
        self.Q_l = self.kwargs.get('Q_l')
        self.Q_c = self.kwargs.get('Q_c')
        self.span = 10e6
        self.nop = 1001
        self.sweep_type = 'LIN'
        self.averages = 1
        self.Average = False
        self.startfreq, self.stopfreq = self.centerfreq + np.array((-1, 1)) * self.span / 2

        self.get_all()
    
    def get_all(self):
        self.get_Average()
        self.get_averages()
        self.get_startfreq()
        self.get_stopfreq()
        self.get_centerfreq()
        self.get_span()
        self.get_nop()
        self.get_freqpoints()
        self.get_sweeptime()
        self.get_sweeptime_averages()


    def avg_clear(self):
        pass

    def avg_status(self):
        return

    def do_set_Average(self, status):
        """
        Set status of Average

        Input:
            status (boolean)

        Output:
            None
        """
        self.Average = status

    def do_get_Average(self):
        """
        Get status of Average

        Input:
            None

        Output:
            Status of Averaging (boolean)
        """
        return self.Average

    def do_set_averages(self, av):
        """
        Set number of averages

        Input:
            av (int) : Number of averages
        Output:
            None
        """
        self.averages = av

    def do_get_averages(self):
        """
        Get number of averages

        Input:
            None
        Output:
            Number of averages
        """
        return self.averages

    def get_freqpoints(self, query=False):
        return np.linspace(self.centerfreq - self.span / 2,
                           self.centerfreq + self.span / 2,
                           self.nop)

    def get_tracedata(self, format='AmpPha'):
        f = self.get_freqpoints()
        S_21 = self.func(f, *self.args, **self.kwargs)
        if format == 'AmpPha':
            return np.abs(S_21), np.angle(S_21)
        elif format == 'RealImag':
            return np.real(S_21), np.imag(S_21)

    def do_set_startfreq(self, val):
        """
        Set Start frequency

        Input:
            span (float) : Frequency in Hz

        Output:
            None
        """
        self.startfreq = val
        
    def do_get_startfreq(self):
        """
        Get Start frequency

        Input:
            None

        Output:
            span (float) : Start Frequency in Hz
        """
        return self.startfreq

    def do_set_stopfreq(self, val):
        """
        Set STop frequency

        Input:
            val (float) : Stop Frequency in Hz

        Output:
            None
        """
        self.stopfreq = val

    def do_get_stopfreq(self):
        """
        Get Stop frequency

        Input:
            None

        Output:
            val (float) : Start Frequency in Hz
        """
        return self.stopfreq

    def do_set_centerfreq(self, cf):
        """
        Set the center frequency

        Input:
            cf (float) :Center Frequency in Hz

        Output:
            None
        """
        self.centerfreq = cf

    def do_get_centerfreq(self):
        """
        Get the center frequency

        Input:
            None

        Output:
            cf (float) :Center Frequency in Hz
        """
        return self.centerfreq

    def do_set_span(self, span):
        """
        Set Span

        Input:
            span (float) : Span in KHz

        Output:
            None
        """
        self.span = span

    def do_get_span(self):
        """
        Get Span

        Input:
            None

        Output:
            span (float) : Span in Hz
        """
        return self.span

    def do_set_nop(self, nop):
        """
        Set Number of Points (nop) for sweep

        Input:
            nop (int) : Number of Points

        Output:
            None
        """
        self.nop = nop

    def do_get_nop(self):
        """
        Get Number of Points (nop) for sweep

        Input:
            None
        Output:
            nop (int)
        """
        return self.nop

    def do_get_sweep_type(self):
        """
        Get the Sweep Type

        Input:
            None

        Output:
            Sweep Type (string). One of
            LIN:    Frequency-based linear sweep
            LOG:    Frequency-based logarithmic sweep
            SEGM:   Segment-based sweep with frequency-based segments
            POW:    Power-based sweep with CW frequency
            CW:     Single frequency mode
        """
        return self.sweep_type

    def do_set_sweep_type(self, swtype):
        """
        Set the Sweep Type
        Input:
            swtype (string):    One of
                LIN:    Frequency-based linear sweep
                LOG:    Frequency-based logarithmic sweep
                SEGM:   Segment-based sweep with frequency-based segments
                POW:    Power-based sweep with CW frequency
                CW:     Time-based sweep with CW frequency

        Output:
            None
        """
        self.sweep_type = swtype

    def do_get_sweeptime_averages(self):
        """
        Get sweeptime

        Input:
            None

        Output:
            sweep time (float) times number of averages: sec
        """
        return 1

    def do_get_sweeptime(self):
        """
        Get sweeptime

        Input:
            None

        Output:
            sweep time (float) : sec
        """
        return 1

    def pre_measurement(self):
        """
        Set everything needed for the measurement
        Averaging has to be enabled.
        Trigger count is set to number of averages
        """
        pass

    def start_measurement(self):
        """
        This function is called at the beginning of each single measurement in the spectroscopy script.
        Here, it starts n sweeps, where n is the active channels trigger count.
        Also, the averages need to be reset.
        """
        pass

    def post_measurement(self):
        """
        Bring the VNA back to a mode where it can be easily used by the operator.
        """
        pass

    def ready(self):
        """
        This is a proxy function, returning True when the VNA is on HOLD after finishing the required number of averages .
        """
        return True




# Yokogawa_GS820.py driver for Yokogawa GS820 multi channel source measure unit
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
import numpy
import struct
import time

class Yokogawa_GS820(Instrument):
    '''
    This is the driver for the Yokogawa GS820 multi channel source measure unit

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Yokogawa_GS820', address='<GBIP address>',
        reset=<bool>)
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the Yokogawa GS820, and communicates with the wrapper.

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
            reset (bool)     : resets to default values

        Output:
            None
        '''
        # Initialize wrapper functions
        logging.info(__name__ + ' : Initializing instrument Yokogawa_GS820')
        Instrument.__init__(self, name, tags=['physical'])

        # Add some global constants
        self._address = address
        self._numchs = 2
        self._visainstrument = visa.instrument(self._address)

        # Add parameters to wrapper
        self.add_parameter('source_range',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            units='', type=str)
        self.add_parameter('sense_range',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            units='', type=str)
        self.add_parameter('source_mode',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=str, units='')
        self.add_parameter('sense_mode',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=str, units='')
        self.add_parameter('source_trig',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=str, units='')
        self.add_parameter('sense_trig',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=str, units='')
        self.add_parameter('source_delay',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=float, minval=15e-6, maxval=3600, units='s')
        self.add_parameter('sense_delay',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=float, minval=15e-6, maxval=3600, units='s')
        self.add_parameter('sweep_mode',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=str, units='')
        self.add_parameter('4W',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=str, units='')
        self.add_parameter('level', 
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1,self._numchs),channel_prefix='ch%d_',
            type=float, units='')
        self.add_parameter('value', flags=Instrument.FLAG_GET,
            channels=(1, self._numchs), channel_prefix='ch%d_',
            type=float, units='')
        self.add_parameter('output',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1,self._numchs),channel_prefix='ch%d_',
            type=str, units='')
        self.add_parameter('sync',
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            type=str, units='')


        # Add functions to wrapper
        self.add_function('reset')
        self.add_function('get_all')
        self.add_function('set_range_auto')
        self.add_function('set_defaults')
        self.add_function('ramp_ch1_current')
        self.add_function('ramp_ch2_current')


        if reset:
            self.reset()
        else:
            self.get_all()
            self.set_defaults()

# functions
    def reset(self):
        '''
        Resets instrument to default values

        Input:
            None
    
        Output:
            None
        '''
        logging.debug(__name__ + ' : Resetting instrument')
        self._visainstrument.write('*RST')
        self.get_all()
      
    def get_all(self):
        '''
        Reads all relevant parameters from instrument

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Get all relevant data from device')
        self.get_sync()
        for ch in range(self._numchs):
            self.get('ch%d_source_range' % (ch+1))
            self.get('ch%d_sense_range' % (ch+1))
            self.get('ch%d_source_mode' % (ch+1))
            self.get('ch%d_sense_mode' % (ch+1))
            self.get('ch%d_source_trig' % (ch+1))
            self.get('ch%d_sense_trig' % (ch+1))
            self.get('ch%d_source_delay' % (ch+1))
            self.get('ch%d_sense_delay' % (ch+1))
            self.get('ch%d_4W' % (ch+1))
            self.get('ch%d_level' % (ch+1))
            self.get('ch%d_output' % (ch+1))

    def set_range_auto(self, channel):
        '''
        Switch autorange on.
        '''
        logging.debug('Set range to auto mode')
        self._visainstrument.write(':chan%d:sens:rang:auto on' % channel)
        
    def set_defaults(self):
        '''
        Set to driver defaults:
        Source range for channel 2 +-18 V
        Source mode for channel 2 voltage mode
        Source mode for channel 1 voltage mode
        Source range for channel 1 +-200 mV
        Sense mode for channel 1 voltage mode
        Sensee range for channel 1 +-200 mV
        Sourse trigger is set to external
        Sense trigger is set connected to source
        Source delay is set to 15 us
        Sense delay is set to 300 ms
        The measuring mode is set to 2-wire sense
        '''
        channel = 1
        self.set('ch%d_4W' % channel, 'off')
        self.set('ch%d_source_mode' % channel, 'volt')
        self.set('ch%d_sense_mode' % channel, 'volt')
        self.set('ch%d_source_range' % channel, '200mV')
        self.set('ch%d_sense_range' % channel, '200mV')
        self.set('ch%d_source_delay' % channel, 15e-6)
        self.set('ch%d_sense_delay' % channel, 0.3)
        self.set('ch%d_source_trig' % channel, 'ext')
        self.set('ch%d_sense_trig' % channel, 'sour')
        self._set_func_par_value(channel, 'sour', 'mode', 'sing')
    
        channel = 2
        self.set('ch%d_4W' % channel, 'off')
        self.set('ch%d_source_mode' % channel, 'volt')
        self.set('ch%d_sense_mode' % channel, 'volt')
        self.set('ch%d_source_range' % channel, '18V')
        self.set('ch%d_sense_range' % channel, '18V')
        self.set('ch%d_source_delay' % channel, 15e-6)
        self.set('ch%d_sense_delay' % channel, 0.3)
        self.set('ch%d_source_trig' % channel, 'ext')
        self.set('ch%d_sense_trig' % channel, 'sour')
        
 # parameters
 
 
    def do_set_sync(self, val):
        '''
        Turns "on"/"off" the interchannel synchronization.
        The first channel is always " The master", second - "the slave"
        '''    
        logging.debug('Sets the channels in synchronized regime')
        if val in ['on','off']:
            logging.debug('Set mode to %s' % val)
            self._visainstrument.write(':sync:chan %s' % val)
        else:
            logging.error('Invalid value %s' % val)
            
    def do_get_sync(self):
        '''
        Gets synchronization mode.
        '''
        logging.debug('Gets synchronization mode')
        ans = self._visainstrument.ask(':sync:chan?')
        
        if int(ans):
            return 'on'
        else:
            return 'off'

    def do_set_source_range(self, val, channel):
        '''
        sets source range to definite channel 
         
        Input:
            val (string)  :
            channel (int) : 

        Output:

        '''
        logging.debug('Set source range to %s' % val)
        self._set_func_par_value(channel, 'sour', 'rang', val)
        
        

    def do_get_source_range(self, channel):
        '''
        Get source range 
        '''
        logging.debug('Get source range')
        return self._get_func_par(channel, 'sour', 'rang')
        
        
    def do_set_sense_range(self, val, channel):
        '''
        Set sense range to the specified channel

        Input:
            val (string)      : 
            channel (int)     : 

        Output:
            None
        '''
        logging.debug('Set sense range to %s' % val)
        if (val == 'auto'):
            self.set_range_auto(channel)
        else:
            self._set_func_par_value(channel, 'sens', 'rang', val)

    def do_get_sense_range(self, channel):
        '''
        Get sense range for the current mode.
        
        Input:
            channel(int): 
        Output:
            range (string) : 
        '''
        logging.debug('Get sense range')
        return self._get_func_par(channel, 'sens', 'rang')
        
    def do_set_source_mode(self, mode, channel):
        '''
        Set source mode to current regime.
        '''    
        if mode in ['curr','volt']:
            logging.debug('Set mode to %s mode', mode)
            self._set_func_par_value(channel, 'sour', 'func', mode)
        else:
            logging.error('invalid mode %s' % mode)
        self.get_all()

    def do_get_source_mode(self, channel):
        '''
        Get source mode 
        
        Input:
            channel (int) : 
        '''
        logging.debug('Get source mode')
        return self._get_func_par(channel, 'sour', 'func')
        
    def do_set_sweep_mode(self, mode, channel):
        '''
        Set source mode to current regime.
        '''    
        if mode in ['fix']:
            logging.debug('Set mode to %s sweep mode', mode)
            self._set_func_par_value(channel, 'sour', 'mode', mode)
        else:
            logging.error('invalid sweep mode %s' % mode)
        self.get_all()

    def do_get_sweep_mode(self, channel):
        '''
        Get source mode 
        
        Input:
            channel (int) : 
        '''
        logging.debug('Get source sweep mode')
        return self._get_func_par(channel, 'sour', 'mode')

    def do_set_sense_mode(self, mode, channel):
        '''
        Set sense mode to current regime.
        '''    
        if mode in ['curr','volt']:
            logging.debug('Set sense mode to %s mode', mode)
            self._set_func_par_value(channel, 'sens', 'func', mode)
        else:
            logging.error('invalid mode %s' % mode)
        self.get_all()
        
    def do_get_sense_mode(self, channel):
        '''
        Get sense mode for the current mode.
        
        Input:
            channel (int) : 
        '''
        logging.debug('Get sense mode')
        return self._get_func_par(channel, 'sens', 'func')
        
    def do_set_source_trig(self, trig, channel):
        '''
        Set source trigger.
        '''    
        if trig in ['ext','sens', 'aux']:
            logging.debug('Set source trig to %s trigger', trig)
            self._set_func_par_value(channel, 'sour', 'trig', trig)
        else:
            logging.error('invalid trig %s' % trig)
        self.get_all()

    def do_get_source_trig(self, channel):
        '''
        Get source trigger
        '''
        logging.debug('Get source trigger')
        return self._get_func_par(channel, 'sour', 'trig')


    def do_set_sense_trig(self, trig, channel):
        '''
        Set sense trigger.
        '''    
        if trig in ['ext','sour', 'aux']:
            logging.debug('Set sense trig to %s trigger', trig)
            self._set_func_par_value(channel, 'sens', 'trig', trig)
        else:
            logging.error('invalid trig %s' % trig)
        self.get_all()

    def do_get_sense_trig(self, channel):
        '''
        Get sense trigger
        
        '''
        logging.debug('Get sense trigger')
        return self._get_func_par(channel, 'sens', 'trig')
        
    def do_set_source_delay(self, val, channel):
        '''
        Set source delay
        '''
        logging.debug('Set source delay to %s' % val)
        self._set_func_par_value(channel, 'sour', 'del', val)

    def do_get_source_delay(self, channel):
        '''
        Get source delay
        '''
        logging.debug('Get source delay')
        return float(self._get_func_par(channel, 'sour','del'))
        
    def do_set_sense_delay(self, val, channel):
        '''
        Set sense delay
        '''
        logging.debug('Set sense delay to %s' % val)
        self._set_func_par_value(channel, 'sens', 'del', val)


    def do_get_sense_delay(self, channel):
        '''
        Get sense delay
        '''
        logging.debug('Get sense delay')
        return float(self._get_func_par(channel, 'sens','del'))
        
    def do_set_4W(self, val, channel):
        '''
        Sets measurement mode to 4-wire or 2-wire mode.
        Value should be on or off. 
        "On" devotes to 4-wire mode. "Off" devotes to 2-wire mode.
        In the instrument appropriate command is ":sens:rem on/off"
        '''
        if val in ['on','off']:
            logging.debug('Set 4W to %s' % val)
            self._set_func_par_value(channel, 'sens', 'rem', val)
        else:
            logging.error('Invalid value %s' % val)

    def do_get_4W(self, channel):
        '''
        Gets measurement mode
        '''
        logging.debug('Get 4-wire measurement mode')
        ans = self._get_func_par(channel, 'sens', 'rem')
        
        if int(ans):
            return 'on'
        else:
            return 'off'
        
    def do_set_level(self, val, channel):
        '''
        Set measuring level
        '''
        logging.debug('Set measuring level')
        if self.get('ch%d_output' % channel, query = False) == 'off':
            self.set('ch%d_output' % channel, 'on')
        self._set_func_par_value(channel, 'sour', 'lev', val)
        self._visainstrument.write(':chan%d:init' % channel)
        self._visainstrument.write(':trig')
        
    def do_get_level(self, channel):
        '''
        Gets source output level
        '''
        logging.debug('Get measuring level')
        return float(self._get_func_par(channel, 'sour','lev'))
        
    def do_set_output(self, val, channel):
        '''
        Sets outputs of channels in "on" or "off" position.
        
        '''
        if val in ['on','off']:
            logging.debug('Set output to %s' % val)
            self._set_func_par_value(channel, 'outp', 'stat', val)
            if (self.get_sync() == 'on')  and (channel == 1):
                self.set_ch2_output(self.get_ch2_output(query = False))
                
        else:
            logging.error('Invalid value %s' % val)
            
    def do_get_output(self, channel):
        '''
        Gets output state
        '''
        logging.debug('Get status of output')
        ans=self._get_func_par(channel, 'outp','stat')
        if ans=='zero':
            return 'off'
        if int(ans):
            return 'on'
        else:
            return 'off'
        
    def do_get_value(self, channel):
        '''
        Gets measured value
        '''
        logging.debug('Get measured value')
        return float(self._visainstrument.ask(':chan%d:fetc?' % channel))

  # core communication
    def _set_func_par_value(self, ch, func, par, val):
        '''
        For internal use only!!
        Changes the value of the parameter for the function specified

        Input:
            ch (int)
            func (string) :
            par (string)  :
            val (depends) :

        Output:
            None
        '''
        string = ':chan%d:%s:%s %s' %(ch, func, par, val)
        logging.debug(__name__ + ' : Set instrument to %s' % string)
        self._visainstrument.write(string)

    def _get_func_par(self, ch, func, par):
        '''
        For internal use only!!
        Reads the value of the parameter for the function specified
        from the instrument

        Input:
            ch (int)      :
            func (string) :
            par (string)  :

        Output:
            val (string) :
        '''
        string = ':chan%d:%s:%s?' %(ch, func, par)
        ans = self._visainstrument.ask(string)
        logging.debug(__name__ + ' : ask instrument for %s (result %s)' % \
            (string, ans))
        return ans.lower()

    def ramp_ch1_current(self,target, step, wait=0.1, showvalue=True):
        start = self.get_ch1_level()
        if(target < start): step = -step
        a = numpy.concatenate( (numpy.arange(start, target, step)[1:], [target]) )
        for i in a:
            if showvalue==True: print i,
            self.set_ch1_level(i)
            time.sleep(wait)


    def ramp_ch2_current(self,target, step, wait=0.1, showvalue=True):
        start = self.get_ch2_level()
        if(target < start): step = -step
        a = numpy.concatenate( (numpy.arange(start, target, step)[1:], [target]) )
        for i in a:
            if showvalue==True: print i,
            self.set_ch2_level(i)
            time.sleep(wait)



# spectroscopy.py spectroscopy measurement class for use with a vector network analyzer
# JB/MP/AS@KIT 04/2015, 08/2015, 01/2016

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import numpy as np
import logging
from time import sleep, time
import sys
import threading

import qkit
if qkit.module_available("matplotlib"):
    import matplotlib.pylab as plt
if qkit.module_available("scipy"):
    from scipy.optimize import curve_fit
    from scipy.interpolate import interp1d, UnivariateSpline
from qkit.storage import store as hdf
from qkit.analysis.resonator import Resonator as resonator
from qkit.gui.plot import plot as qviewkit
from qkit.gui.notebook.Progress_Bar import Progress_Bar
from qkit.measure.measurement_class import Measurement
import qkit.measure.write_additional_files as waf


##################################################################

class spectrum(object):
    """
    Class for spectroscopy measurements with a VNA
    usage:
    m = spectrum(vna = vna1)

    m.set_x_parameters(arange(-0.05,0.05,0.01),'flux coil current',coil.set_current, unit = 'mA')  outer scan in 3D
    m.set_y_parameters(arange(4e9,7e9,10e6),'excitation frequency',mw_src1.set_frequency, unit = 'Hz')

    m.landscape.generate_fit_function_xy(...) for 3D scan, can be called several times and appends the current landscape
    m.landscape.generate_fit_function_xz(...) for 2D or 3D scan, adjusts the vna freqs with respect to x

    m.measure_XX()
    """

    def __init__(self, vna, exp_name='', sample=None):

        self.vna = vna
        self.averaging_start_ready = "start_measurement" in self.vna.get_function_names() and "ready" in self.vna.get_function_names()
        if not self.averaging_start_ready: logging.warning(
            __name__ + ': With your VNA instrument driver (' + self.vna.get_type() + '), I can not see when a measurement is complete. So I only wait for a specific time and hope the VNA has finished. Please consider implemeting the necessary functions into your driver.')
        self.exp_name = exp_name
        self._sample = sample
        self.landscape = Landscape(vna=vna,spec=self)
        self.tdx = 0.002  # [s]
        self.tdy = 0.002  # [s]

        self.vna_poll_interval = 0.1 # interval in seconds in which the vna is queried to be ready.

        self.comment = ''
        self.dirname = None

        self.x_set_obj = None
        self.y_set_obj = None
        self.x_vec = None
        self.y_vec = None

        self.progress_bar = True
        self._fit_resonator = False
        self._plot_comment = ""

        self.set_log_function()
        self.set_log_function_2D()

        self.open_qviewkit = True
        self.qviewkit_singleInstance = False

        self._measurement_object = Measurement()
        self._measurement_object.measurement_type = 'spectroscopy'
        self._measurement_object.sample = self._sample
        self.measurement_object_axis_name = 'frequency'

        self._qvk_process = False
        self._scan_dim = None
        self._scan_time = False

    def set_log_function(self, func=None, name=None, unit=None, log_dtype=None):
        '''
        A function (object) can be passed to the measurement loop which is excecuted before every x iteration
        but after executing the x_object setter in 2D measurements and before every line (but after setting
        the x value) in 3D measurements.
        The return value of the function of type float or similar is stored in a value vector in the h5 file.

        Call without any arguments to delete all log functions. The timestamp is automatically saved.

        func: function object in list form
        name: name of logging parameter appearing in h5 file, default: 'log_param'
        unit: unit of logging parameter, default: ''
        log_dtype: h5 data type, default: 'f' (float32)
        '''

        if name == None:
            try:
                name = ['log_param'] * len(func)
            except Exception:
                name = None
        if unit == None:
            try:
                unit = [''] * len(func)
            except Exception:
                unit = None
        if log_dtype == None:
            try:
                log_dtype = ['f'] * len(func)
            except Exception:
                log_dtype = None

        self.log_function = []
        self.log_name = []
        self.log_unit = []
        self.log_dtype = []

        if func != None:
            for i, f in enumerate(func):
                self.log_function.append(f)
                self.log_name.append(name[i])
                self.log_unit.append(unit[i])
                self.log_dtype.append(log_dtype[i])
    
    def set_log_function_2D(self, func=None, name=None, unit=None, y=None, y_name=None, y_unit=None, log_dtype=None):
        '''
        A function (object) can be passed to the measurement loop which is excecuted before every x iteration
        but after executing the x_object setter in 2D measurements and before every line (but after setting
        the x value) in 3D measurements.
        The return values of the function of type 1D-list or similar is stored in a value matrix in the h5 file.

        Call without any arguments to delete all log functions. The timestamp is automatically saved.

        func: function object in list form, returning a list each
        name: name of logging parameter appearing in h5 file, default: 'log_param'
        unit: unit of logging parameter, default: ''
        log_dtype: h5 data type, default: 'f' (float32)
        '''
        if name == None:
            try:
                name = ['log_param'] * len(func)
            except Exception:
                name = None
        if unit == None:
            try:
                unit = [''] * len(func)
            except Exception:
                unit = None
        if log_dtype == None:
            try:
                log_dtype = ['f'] * len(func)
            except Exception:
                log_dtype = None

        self.log_function_2D = []
        self.log_name_2D = []
        self.log_unit_2D = []
        self.log_y_2D = []
        self.log_y_name_2D = []
        self.log_y_unit_2D = []
        self.log_dtype_2D = []

        if func != None:
            for i, _ in enumerate(func):
                self.log_function_2D.append(func[i])
                self.log_name_2D.append(name[i])
                self.log_unit_2D.append(unit[i])
                self.log_dtype_2D.append(log_dtype[i])
                self.log_y_2D.append(y[i])
                self.log_y_name_2D.append(y_name[i])
                self.log_y_unit_2D.append(y_unit[i])

    def set_x_parameters(self, x_vec, x_coordname, x_set_obj, x_unit=""):
        """
        Sets parameters for sweep. In a 3D measurement, the x-parameters will be the "outer" sweep.
        For every x value all y values are swept
        Input:
        x_vec (array): conains the sweeping values
        x_coordname (string)
        x_instrument (obj): callable object to execute with x_vec-values (i.e. vna.set_power())
        x_unit (string): optional
        """
        self.x_vec = x_vec
        self.x_coordname = x_coordname
        self.x_unit = x_unit
        self.x_set_obj = x_set_obj
        if self.landscape.xylandscapes:
            self.landscape.delete_landscape_function_xy()
            logging.warning('xy landscape has been deleted')
        if self.landscape.xzlandscape_func:
            self.landscape.delete_landscape_function_xz()
            logging.warning('xz landscape has been deleted')

    def set_y_parameters(self, y_vec, y_coordname, y_set_obj, y_unit="", delete_landscape=True):
        """
        Sets parameters for sweep. In a 3D measurement, the x-parameters will be the "outer" sweep.
        For every x value all y values are swept
        Input:
        y_vec (array): contains the sweeping values
        y_coordname (string)
        y_instrument (obj): callable object to execute with x_vec-values (i.e. vna.set_power())
        y_unit (string): optional
        delete_landscape: True (default) if previously generated xz-landscape should be deleted
        """
        self.y_vec = y_vec
        self.y_coordname = y_coordname
        self.y_set_obj = y_set_obj
        self.y_unit = y_unit
        if self.landscape.xylandscapes:
            self.landscape.delete_landscape_function_xy()
            logging.warning('xy landscape has been deleted')
        if self.landscape.xzlandscape_func and delete_landscape:
            self.landscape.delete_landscape_function_xz()
            logging.warning('xz landscape has been deleted. You can switch this off if you whish.')

    def _prepare_measurement_vna(self):
        '''
        all the relevant settings from the vna are updated and called
        '''

        self.vna.get_all()
        # ttip.get_temperature()
        self._nop = self.vna.get_nop()
        self._sweeptime_averages = self.vna.get_sweeptime_averages()
        if self._scan_dim == 1 or not self.landscape.xzlandscape_func: # normal scan
            self._freqpoints = self.vna.get_freqpoints()
        else:
            self._freqpoints = self.landscape.get_freqpoints_xz()
        if self.averaging_start_ready: self.vna.pre_measurement()

    def _prepare_measurement_file(self):
        '''
        creates the output .h5-file with distinct dataset structures for each measurement type.
        at this point all measurement parameters are known and put in the output file
        '''

        self._data_file = hdf.Data(name=self._file_name, mode='a')
        self._measurement_object.uuid = self._data_file._uuid
        self._measurement_object.hdf_relpath = self._data_file._relpath
        self._measurement_object.instruments = qkit.instruments.get_instrument_names()

        self._measurement_object.save()
        self._mo = self._data_file.add_textlist('measurement')
        self._mo.append(self._measurement_object.get_JSON())

        # write logfile and instrument settings
        self._write_settings_dataset()
        self._log = waf.open_log_file(self._data_file.get_filepath())

        self._data_freq = self._data_file.add_coordinate('frequency', unit='Hz')
        if not self._scan_time:
            self._data_freq.add(self._freqpoints)
            sweep_vector = self._data_freq
        else:
            self._data_freq.add([self.vna.get_centerfreq()])
            self._data_time = self._data_file.add_coordinate('time', unit='s')
            self._data_time.add(np.arange(0, self._nop, 1) * self.vna.get_sweeptime() / (self._nop - 1))
            sweep_vector = self._data_time

        if self._scan_dim == 1:
            self._data_real = self._data_file.add_value_vector('real', x=sweep_vector, unit='', save_timestamp=True)
            self._data_imag = self._data_file.add_value_vector('imag', x=sweep_vector, unit='', save_timestamp=True)
            self._data_amp = self._data_file.add_value_vector('amplitude', x=sweep_vector, unit='arb. unit',
                                                              save_timestamp=True)
            self._data_pha = self._data_file.add_value_vector('phase', x=sweep_vector, unit='rad',
                                                              save_timestamp=True)

        if self._scan_dim == 2:
            self._data_x = self._data_file.add_coordinate(self.x_coordname, unit=self.x_unit)
            self._data_x.add(self.x_vec)
            self._data_amp = self._data_file.add_value_matrix('amplitude', x=self._data_x, y=sweep_vector,
                                                              unit='arb. unit', save_timestamp=True)
            self._data_pha = self._data_file.add_value_matrix('phase', x=self._data_x, y=sweep_vector, unit='rad',
                                                              save_timestamp=True)

            if self.log_function != None:  # use logging
                self._log_value = []
                for i in range(len(self.log_function)):
                    self._log_value.append(
                        self._data_file.add_value_vector(self.log_name[i], x=self._data_x, unit=self.log_unit[i],
                                                         dtype=self.log_dtype[i]))

            if self._nop < 10:
                """creates view: plot middle point vs x-parameter, for qubit measurements"""
                self._views = [
                    self._data_file.add_view("amplitude_midpoint",x=self._data_x,y=self._data_amp,
                                             view_params=dict(transpose=True,default_trace=self._nop // 2,linecolors=[(200,200,100)])),
                    self._data_file.add_view("phase_midpoint", x=self._data_x, y=self._data_pha,
                                             view_params=dict(transpose=True, default_trace=self._nop // 2, linecolors=[(200, 200, 100)]))
                ]

        if self._scan_dim == 3:
            self._data_x = self._data_file.add_coordinate(self.x_coordname, unit=self.x_unit)
            self._data_x.add(self.x_vec)
            self._data_y = self._data_file.add_coordinate(self.y_coordname, unit=self.y_unit)
            self._data_y.add(self.y_vec)

            if self._nop == 0:  # saving in a 2D matrix instead of a 3D box HR: does not work yet !!! test things before you put them online.
                self._data_amp = self._data_file.add_value_matrix('amplitude', x=self._data_x, y=self._data_y,
                                                                  unit='arb. unit', save_timestamp=False)
                self._data_pha = self._data_file.add_value_matrix('phase', x=self._data_x, y=self._data_y, unit='rad',
                                                                  save_timestamp=False)
            else:
                self._data_amp = self._data_file.add_value_box('amplitude', x=self._data_x, y=self._data_y,
                                                               z=sweep_vector, unit='arb. unit',
                                                               save_timestamp=False)
                self._data_pha = self._data_file.add_value_box('phase', x=self._data_x, y=self._data_y,
                                                               z=sweep_vector, unit='rad', save_timestamp=False)

            if self.log_function != None:  # use logging
                self._log_value = []
                for i in range(len(self.log_function)):
                    self._log_value.append(
                        self._data_file.add_value_vector(self.log_name[i], x=self._data_x, unit=self.log_unit[i],
                                                         dtype=self.log_dtype[i]))

            if self.log_function_2D != None:  # use 2D logging
                self._log_y_value_2D = []
                for i in range(len(self.log_y_name_2D)):
                    if self.log_y_name_2D[i] not in self.log_y_name_2D[:i]:  # add y coordinate for 2D logging
                        self._log_y_value_2D.append(self._data_file.add_coordinate(self.log_y_name_2D[i], unit=self.log_y_unit_2D[i], folder='data'))  # possibly use "data1"
                        self._log_y_value_2D[i].add(self.log_y_2D[i])
                    else:  # use y coordinate for 2D logging if it is already added
                        self._log_y_value_2D.append(self._log_y_value_2D[np.squeeze(np.argwhere(self.log_y_name_2D[i] == np.array(self.log_y_name_2D[:i])))])
                
                self._log_value_2D = []
                for i in range(len(self.log_function_2D)):
                    self._log_value_2D.append(
                        self._data_file.add_value_matrix(self.log_name_2D[i], x=self._data_x, y=self._log_y_value_2D[i], unit=self.log_unit_2D[i],
                                                         dtype=self.log_dtype_2D[i], folder='data'))  # possibly use "data1"

        if self.comment:
            self._data_file.add_comment(self.comment)

        if self.qviewkit_singleInstance and self.open_qviewkit and self._qvk_process:
            self._qvk_process.terminate()  # terminate an old qviewkit instance

    def _write_settings_dataset(self):
        self._settings = self._data_file.add_textlist('settings')
        settings = waf.get_instrument_settings(self._data_file.get_filepath())
        self._settings.append(settings)

    def measure_1D(self, rescan=True, web_visible=True):
        '''
        measure method to record a single (averaged) VNA trace, S11 or S21 according to the setting on the VNA
        rescan: If True (default), the averages on the VNA are cleared and a new measurement is started.
                If False, it will directly take the data from the VNA without waiting.
        '''
        self._scan_dim = 1

        self._measurement_object.measurement_func = 'measure_1D'
        self._measurement_object.x_axis = self.measurement_object_axis_name
        self._measurement_object.y_axis = ''
        self._measurement_object.z_axis = ''
        self._measurement_object.web_visible = web_visible

        if not self.dirname:
            self.dirname = 'VNA_tracedata'
        self._file_name = self.dirname.replace(' ', '').replace(',', '_')
        if self.exp_name:
            self._file_name += '_' + self.exp_name
        self._prepare_measurement_vna()
        self._prepare_measurement_file()

        """opens qviewkit to plot measurement, amp and pha are opened by default"""
        if self.open_qviewkit:
            self._qvk_process = qviewkit.plot(self._data_file.get_filepath(), datasets=['amplitude', 'phase'])
        if self._fit_resonator:
            self._resonator = resonator(self._data_file.get_filepath())

        qkit.flow.start()
        if rescan:
            if self.averaging_start_ready:
                self.vna.start_measurement()
                ti = time()
                if self.progress_bar:
                    self._p = Progress_Bar(self.vna.get_averages(), self.dirname, self.vna.get_sweeptime())
                qkit.flow.sleep(.2)
                while not self.vna.ready():
                    if time() - ti > self.vna.get_sweeptime(query=False):
                        if self.progress_bar: self._p.iterate()
                        ti = time()
                    qkit.flow.sleep(.2)
                if self.progress_bar:
                    while self._p.progr < self._p.max_it:
                        self._p.iterate()
            else:
                self.vna.avg_clear()
                if self.vna.get_averages() == 1 or self.vna.get_Average() == False:  # no averaging
                    if self.progress_bar: self._p = Progress_Bar(1, self.dirname, self.vna.get_sweeptime())
                    qkit.flow.sleep(self.vna.get_sweeptime())  # wait single sweep
                    if self.progress_bar: self._p.iterate()
                else:  # with averaging
                    if self.progress_bar:
                        self._p = Progress_Bar(self.vna.get_averages(), self.dirname, self.vna.get_sweeptime())
                    if "avg_status" in self.vna.get_function_names():
                        for a in range(self.vna.get_averages()):
                            while self.vna.avg_status() <= a:
                                qkit.flow.sleep(.2)  # maybe one would like to adjust this at a later point
                            if self.progress_bar: self._p.iterate()
                    else:  # old style
                        for a in range(self.vna.get_averages()):
                            qkit.flow.sleep(self.vna.get_sweeptime())  # wait single sweep time
                            if self.progress_bar: self._p.iterate()

        data_amp, data_pha = self.vna.get_tracedata()
        data_real, data_imag = self.vna.get_tracedata('RealImag')

        self._data_amp.append(data_amp)
        self._data_pha.append(data_pha)
        self._data_real.append(data_real)
        self._data_imag.append(data_imag)
        if self._fit_resonator:
            self._do_fit_resonator()

        qkit.flow.end()
        self._end_measurement()

    def measure_2D(self, web_visible=True):
        '''
        measure method to record a (averaged) VNA trace, S11 or S21 according to the setting on the VNA
        for all parameters x_vec in x_obj
        '''

        if not self.x_set_obj:
            logging.error('axes parameters not properly set...aborting')
            return
        if self.landscape.xzlandscape_func:  # The vna limits need to be adjusted, happens in the frequency wrapper
            self.x_set_obj = self.landscape.vna_frequency_wrapper(self.x_set_obj)
        if len(self.x_vec) == 0:
            logging.error('No points to measure given. Check your x vector... aborting')
            return
        self._scan_dim = 2

        self._measurement_object.measurement_func = 'measure_2D'
        self._measurement_object.x_axis = self.x_coordname
        self._measurement_object.y_axis = self.measurement_object_axis_name
        self._measurement_object.z_axis = ''
        self._measurement_object.web_visible = web_visible

        if not self.dirname:
            self.dirname = self.x_coordname
        self._file_name = '2D_' + self.dirname.replace(' ', '').replace(',', '_')
        if self.exp_name:
            self._file_name += '_' + self.exp_name

        if self.progress_bar: self._p = Progress_Bar(len(self.x_vec), '2D VNA sweep ' + self.dirname,
                                                     self.vna.get_sweeptime_averages())

        self._prepare_measurement_vna()
        self._prepare_measurement_file()

        """opens qviewkit to plot measurement, amp and pha are opened by default"""
        if self._nop < 10:
            self._data_file.hf.hf.attrs['default_ds'] =['views/amplitude_midpoint', 'views/phase_midpoint']
        else:
            self._data_file.hf.hf.attrs['default_ds'] = ['data0/amplitude_midpoint', 'data0/phase_midpoint']
        
        if self.open_qviewkit:
            self._qvk_process = qviewkit.plot(self._data_file.get_filepath(),datasets=list(self._data_file.hf.hf.attrs['default_ds']))
        if self._fit_resonator:
            self._resonator = resonator(self._data_file.get_filepath())
        self._measure()

    def measure_3D(self, web_visible=True):
        '''
        measure full window of vna while sweeping x_set_obj and y_set_obj with parameters x_vec/y_vec. sweep over y_set_obj is the inner loop, for every value x_vec[i] all values y_vec are measured.

        optional: measure method to perform the measurement according to landscape, if set
        self.span is the range (in units of the vertical plot axis) data is taken around the specified funtion(s)
        note: make sure to have properly set x,y vectors before generating traces
        '''
        if not self.x_set_obj or not self.y_set_obj:
            logging.error('axes parameters not properly set...aborting')
            return
        if len(self.x_vec) * len(self.y_vec) == 0:
            logging.error('No points to measure given. Check your x ad y vector... aborting')
            return
        if self.landscape.xzlandscape_func:  # The vna limits need to be adjusted, happens in the frequency wrapper
            self.x_set_obj = self.landscape.vna_frequency_wrapper(self.x_set_obj)

        self._scan_dim = 3

        self._measurement_object.measurement_func = 'measure_3D'
        self._measurement_object.x_axis = self.x_coordname
        self._measurement_object.y_axis = self.y_coordname
        self._measurement_object.z_axis = self.measurement_object_axis_name
        self._measurement_object.web_visible = web_visible

        if not self.dirname:
            self.dirname = self.x_coordname + ', ' + self.y_coordname
        self._file_name = '3D_' + self.dirname.replace(' ', '').replace(',', '_')
        if self.exp_name:
            self._file_name += '_' + self.exp_name

        self._prepare_measurement_vna()
        self._prepare_measurement_file()
        """opens qviewkit to plot measurement, amp and pha are opened by default"""
        """only middle point in freq array is plotted vs x and y"""
        if self.open_qviewkit: self._qvk_process = qviewkit.plot(self._data_file.get_filepath(),
                                                                 datasets=['amplitude', 'phase'])
        if self._fit_resonator:
            self._resonator = resonator(self._data_file.get_filepath())

        if self.progress_bar:
            if self.landscape.xylandscapes:
                truth = np.full((len(self.y_vec), len(self.x_vec)), False)  # first, nothing is selected:
                for e in self.landscape.xylandscapes:
                    if not e['blacklist']:
                        truth = np.logical_or(truth,
                                              (np.abs(self.y_vec[:, np.newaxis] - e['center_points']) <= e['y_span'] / 2) *  # check y span
                                              (e['x_range'][0] <= self.x_vec) * (self.x_vec <= e['x_range'][1])) # check x range
    
                for e in self.landscape.xylandscapes:
                    if e['blacklist']:  # exclude blacklisted areas
                        truth = np.logical_and(truth,
                                               np.logical_not(
                                                  (np.abs(self.y_vec[:, np.newaxis] - e['center_points']) <= e['y_span'] / 2) *
                                                  (e['x_range'][0] <= self.x_vec) * (self.x_vec <= e['x_range'][1]))
                                               )
                points = np.sum(truth)
            else:
                points = len(self.x_vec) * len(self.y_vec)
            self._p = Progress_Bar(points, '3D VNA sweep ' + self.dirname, self.vna.get_sweeptime_averages())

        self._measure()

    def set_cw_time_measurement(self, status):
        """
        Activates or deactivates timetrace measurements. Sets the VNA in constant cw mode to measure over time
        and adjusts the sweep vector name for the measurement object
        :param status: (bool) True = timetrace active, False = timetrace inactive
        :param web_visible:
        :return:
        """

        if status:
            self.vna.set_sweep_type('CW')
            self._scan_time = True
            self.measurement_object_axis_name = 'time'

        if status is False:
            self.vna.set_sweep_type('LIN')
            self._scan_time = False
            self.measurement_object_axis_name = 'frequency'

    def _measure(self):
        '''
        measures and plots the data depending on the measurement type.
        the measurement loops feature the setting of the objects and saving the data in the .h5 file.
        '''
        qkit.flow.start()
        try:
            """
            loop: x_obj with parameters from x_vec
            """
            for ix, x in enumerate(self.x_vec):
                self.x_set_obj(x)
                sleep(self.tdx)

                if self.log_function != None:
                    for i, f in enumerate(self.log_function):
                        self._log_value[i].append(float(f()))

                if self.log_function_2D != None:
                    for i, f in enumerate(self.log_function_2D):
                        self._log_value_2D[i].append(f())

                if self._scan_dim == 3:
                    for y in self.y_vec:
                        # loop: y_obj with parameters from y_vec (only 3D measurement)
                        if self.landscape.xylandscapes and not self.landscape.perform_measurement_at_point(x, y, ix):
                            # if point is not of interest (not close to one of the functions)
                            data_amp = np.full(int(self._nop), np.NaN, dtype=np.float16)
                            data_pha = np.full(int(self._nop), np.NaN, dtype=np.float16)  # fill with NaNs
                        else:
                            self.y_set_obj(y)
                            sleep(self.tdy)
                            if self.averaging_start_ready:
                                self.vna.start_measurement()
                                # Check if the VNA is STILL in ready state, then add some delay.
                                # If you manually decrease the poll_inveral, I guess you know what you are doing and will disable this safety query.
                                if self.vna_poll_interval >= 0.1 and self.vna.ready():
                                    logging.debug("VNA STILL ready... Adding delay")
                                    qkit.flow.sleep(
                                        .2)  # just to make sure, the ready command does not *still* show ready

                                while not self.vna.ready():
                                    qkit.flow.sleep(min(self.vna.get_sweeptime_averages(query=False) / 11., self.vna_poll_interval))
                            else:
                                self.vna.avg_clear()
                                qkit.flow.sleep(self._sweeptime_averages)

                            # if "avg_status" in self.vna.get_function_names():
                            #       while self.vna.avg_status() < self.vna.get_averages():
                            #            qkit.flow.sleep(.2) #maybe one would like to adjust this at a later point

                            """ measurement """
                            if not self.landscape.xzlandscape_func:  # normal scan
                                data_amp, data_pha = self.vna.get_tracedata()
                            else:
                                data_amp, data_pha = self.landscape.get_tracedata_xz(x)
                            if self.progress_bar:
                                self._p.iterate()

                        if self._nop == 0:  # this does not work yet.
                            print(data_amp[0], data_amp, self._nop)
                            self._data_amp.append(data_amp[0])
                            self._data_pha.append(data_pha[0])
                        else:
                            self._data_amp.append(data_amp)
                            self._data_pha.append(data_pha)
                        if self._fit_resonator:
                            self._do_fit_resonator()
                        qkit.flow.sleep()
                    """
                    filling of value-box is done here.
                    after every y-loop the data is stored the next 2d structure
                    """
                    self._data_amp.next_matrix()
                    self._data_pha.next_matrix()

                if self._scan_dim == 2:
                    if self.averaging_start_ready:
                        self.vna.start_measurement()
                        if self.vna.ready():
                            logging.debug("VNA STILL ready... Adding delay")
                            qkit.flow.sleep(.2)  # just to make sure, the ready command does not *still* show ready

                        while not self.vna.ready():
                            qkit.flow.sleep(min(self.vna.get_sweeptime_averages(query=False) / 11., .2))
                    else:
                        self.vna.avg_clear()
                        qkit.flow.sleep(self._sweeptime_averages)
                    """ measurement """
                    if not self.landscape.xzlandscape_func:  # normal scan
                        data_amp, data_pha = self.vna.get_tracedata()
                    else:
                        data_amp, data_pha = self.landscape.get_tracedata_xz(x)
                    self._data_amp.append(data_amp)
                    self._data_pha.append(data_pha)

                    if self._fit_resonator:
                        self._do_fit_resonator()
                    if self.progress_bar:
                        self._p.iterate()
                    qkit.flow.sleep()
        finally:
            self._end_measurement()
            qkit.flow.end()

    def _end_measurement(self):
        '''
        the data file is closed and filepath is printed
        '''
        print(self._data_file.get_filepath())
        # qviewkit.save_plots(self._data_file.get_filepath(),comment=self._plot_comment) #old version where we have to wait for the plots
        t = threading.Thread(target=qviewkit.save_plots, args=[self._data_file.get_filepath(), self._plot_comment])
        t.start()
        self._data_file.close_file()
        waf.close_log_file(self._log)
        self.dirname = None
        if self.averaging_start_ready: self.vna.post_measurement()

    def set_resonator_fit(self, fit_resonator=True, fit_function='', f_min=None, f_max=None):
        '''
        sets fit parameter for resonator

        fit_resonator (bool): True or False, default: True (optional)
        fit_function (string): function which will be fitted to the data (optional)
        f_min (float): lower frequency boundary for the fitting function, default: None (optional)
        f_max (float): upper frequency boundary for the fitting function, default: None (optional)
        fit types: 'lorentzian','skewed_lorentzian','circle_fit_reflection', 'circle_fit_notch','fano'
        '''
        if not fit_resonator:
            self._fit_resonator = False
            return
        self._functions = {'lorentzian': 0, 'skewed_lorentzian': 1, 'circle_fit_reflection': 2, 'circle_fit_notch': 3,
                           'fano': 4, 'all_fits': 5}
        try:
            self._fit_function = self._functions[fit_function]
        except KeyError:
            logging.error(
                'Fit function not properly set. Must be either \'lorentzian\', \'skewed_lorentzian\', \'circle_fit_reflection\', \'circle_fit_notch\', \'fano\', or \'all_fits\'.')
        else:
            self._fit_resonator = True
            self._f_min = f_min
            self._f_max = f_max

    def _do_fit_resonator(self):
        '''
        calls fit function in resonator class
        fit function is specified in self.set_fit, with boundaries f_mim and f_max
        only the last 'slice' of data is fitted, since we fit live while measuring.
        '''

        if self._fit_function == 0:  # lorentzian
            self._resonator.fit_lorentzian(f_min=self._f_min, f_max=self._f_max)
        elif self._fit_function == 1:  # skewed_lorentzian
            self._resonator.fit_skewed_lorentzian(f_min=self._f_min, f_max=self._f_max)
        elif self._fit_function == 2:  # circle_reflection
            self._resonator.fit_circle(reflection=True, f_min=self._f_min, f_max=self._f_max)
        elif self._fit_function == 3:  # circle_notch
            self._resonator.fit_circle(notch=True, f_min=self._f_min, f_max=self._f_max)
        elif self._fit_function == 4:  # fano
            self._resonator.fit_fano(f_min=self._f_min, f_max=self._f_max)
        elif self._fit_function == 5: #all fits
            logging.warning("Please performe fits individually, fit all is currently not supported.")
        # self._resonator.fit_all_fits(f_min=self._f_min, f_max = self._f_max)
        else:
            logging.error("Fit function set in spectrum.set_resonator_fit is not supported. Must be either \'lorentzian\', \'skewed_lorentzian\', \'circle_fit_reflection\', \'circle_fit_notch\', \'fano\', or \'all_fits\'.")

    def set_tdx(self, tdx):
        self.tdx = tdx

    def set_tdy(self, tdy):
        self.tdy = tdy

    def get_tdx(self):
        return self.tdx

    def get_tdy(self):
        return self.tdy

    def set_plot_comment(self, comment):
        '''
        Small comment to add at the end of plot pics for more information i.e. good for wiki entries.
        '''
        self._plot_comment = comment

    def gen_fit_function(self, curve_f, curve_p, x_range=None, span=None, p0=[-1, 0.1, 7], units=''):
        """
        Generates an xy landscape fit for 3D scan. Function is deprecated and calls landscape.generate_fit_function
        curve_f: 'spline', '2pt' 'lin, 'parab', 'hyp', specifies the fit function to be employed
        curve_p: set of points that are the basis for the fit in the format [[x1,x2,x3,...],[y1,y2,y3,...]], frequencies in Hz
        span: specify span for the generated fit_function, if not given default of self.span is used
        x_range: specify x_range for the function in the format [x_min, x_max]
        p0 (optional): start parameters for the fit, must be an 1D array of length 3 ([a,b,c,d]),
           where for the parabula p0[3] will be ignored
        units: set this to 'Hz' in order to avoid large values that cause the fit routine to diverge
        The parabolic function takes the form  y = a*(x-b)**2 + c , where (a,b,c) = p0
        The hyperbolic function takes the form y = sqrt[ a*(x-b)**2 + c ], where (a,b,c) = p0

        adds a trace to landscape
        """
        logging.warning("This function is deprecated. Better to use landscape.generate_fit_function_xy")
        self.landscape.generate_fit_function_xy(curve_f=curve_f, curve_p=curve_p, x_range=x_range, y_span=span, p0=p0, units=units)

    def delete_fit_function(self, n=None):
        """
        legacy support for delete_xy_landscape_function. Please use the other function in the future
        delete single fit function n (with 0 being the first one generated) or the complete landscape for n not specified
        :param n: index of single fit function to be deleted
        :return:
        """
        logging.warning('This function is deprecated for clarity purposes.'
                        ' Please use delete_xy_landscape_function in the future')
        self.landscape.delete_landscape_function_xy(n=n)

    def plot_fit_function(self):
        """
        legeacy support for landscape.plot_fit_function_xy. Please use the other function in the future.
        Plots the landscape
        :return:
        """
        self.landscape.plot_xy_landscape()

    def measure_timetrace(self):
        logging.error("This function does no longer exist. Use set_cw_time_mesurement() "
                      "followed by the standard measure functions")


class Landscape:
    """
    Class for landscape scans in combination with spectroscopy. spectrum gnerates an instance of Landscape in __init__
    Use this instance to generate your fit_functions/landscape for your measurement.
    Instance stores the landscape(functions), lets you plot them, checks if points are to be measured and adjusts
    the vna frequencies.
    """
    def __init__(self, vna, spec):
        self.vna = vna
        self.spec = spec # The spectroscopy object
        self.xylandscapes = []  # List containing dicts
        self.xzlandscape_func = None
        self.xz_freqpoints = None
        self.y_span_default = 200e6  # this is for the xy landscape scan, i.e., span of your y_parameter, e.g, mw_frequency
        self.z_span = self.vna.get_span()  # This is for the xz landscape scan i.e. span of vna is adjusted w/ resp to x

    def generate_fit_function_xy(self, curve_f, curve_p, x_range=None, y_span=None, blacklist=False, p0=[-1, 0.1, 7], units=''):
        """
        Use this function if you keep your vna span fixed but want to add a landscape to your x and y parameter.
        I.e., for given x parameter not all y parameters are swept but only those that lie within the specified y-span
        around the fit function.
        Adds a trace to landscape
        curve_f: 'spline', 'lin_spline' 'lin, 'parab', 'hyp', 'transmon' specifies the fit function to be employed
        curve_p: set of points that are the basis for the fit in the format [[x1,x2,x3,...],[y1,y2,y3,...]], frequencies in Hz
        x_range: specify x_range for the function in the format [x_min, x_max]
        y_span: specify the span of your y_axis for the generated fit_function,
                if not given default of self.span is used, if negative the span will be excluded from the measurement
                interesting for overlaping regions
        blacklist: Use this parameter to explicitly exclude a section from the measurement (previously done by negative span)
        p0 (optional): start parameters for the fit, must be an 1D array of length 3 ([a,b,c,d]),
           where for the parabula p0[3] will be ignored
        units: set this to 'Hz' in order to avoid large values that cause the fit routine to diverge

        The parabolic function takes the form  y = a*(x-b)**2 + c , where (a,b,c) = p0
        The hyperbolic function takes the form y = sqrt[ a*(x-b)**2 + c ], where (a,b,c) = p0

        """
        if not qkit.module_available("scipy"):
            raise ImportError('scipy not available.')

        if units == 'Hz':
            multiplier = 1e9
        else:
            multiplier = 1
        x_fit = np.array(curve_p[0])
        y_fit = np.array(curve_p[1])
        if x_range is None:
            x_range = [self.spec.x_vec[0], self.spec.x_vec[-1]]
        if y_span is None:
            y_span = self.y_span_default
        if y_span < 0:
            y_span = -y_span
            blacklist = True
            logging.warning('Using a negative span in generate_fit_function_xy is deprecated. Please use blacklist = True instead for more clarity.')

        try:
            if curve_f == 'lin_spline':
                f = interp1d(x_fit, y_fit, fill_value="extrapolate")
                center_points = f(self.spec.x_vec)
            elif curve_f == 'spline':
                center_points =  UnivariateSpline(x_fit, y_fit)(self.spec.x_vec)
            else:  # curve_fit procedure differs from splines
                if curve_f == 'parab':
                    fit_fct = self.f_parab
                elif curve_f == 'hyp':
                    fit_fct = self.f_hyp
                elif curve_f == 'lin':
                    fit_fct = self.f_lin
                    p0 = p0[:2]
                elif curve_f == 'transmon':
                    fit_fct = self.f_transmon
                else:
                    raise ValueError('function type not known...aborting')
                popt, pcov = curve_fit(fit_fct, x_fit, y_fit / multiplier, p0=p0)
                center_points = multiplier * fit_fct(self.spec.x_vec, *popt)
        except Exception as message:
            print('fit not successful:', message)
        else:
            self.xylandscapes.append({
                'center_points':center_points,
                'y_span':y_span,
                'x_range':[np.min(x_range), np.max(x_range)],
                'blacklist':blacklist
                })
        
    def generate_fit_function_xz(self, curve_f, curve_p, z_span=None, p0=[-1, 0.1, 7]):
        """
        Use this function if you want to adjust your vna span with respect to your x-values, useful if you want to
        track your fresonator, e.g., shifting resonance due qubit resonator coupling or fmr measurements. It works
        in a 2d and 3d scan.
        :param curve_f: fit_function (str) such as 'lin_spline', 'spline', 'parab', 'hyp', 'lin'
        :param curve_p: Valeus to fit in the form [[x0,x1,..xn],[z0,z1,..zn]]
        :param z_span: If None, span from VNA is used. If specified, vna span will be updated.
        :param p0: intial fit parameters
        :return:
        """
        if not qkit.module_available("scipy"):
            raise ImportError('scipy not available.')
        
        if curve_p[1][-1] > 1e6:  # test if z_values are given in Hz or GHz
            multiplier = 1e9
        else:
            multiplier = 1
        if z_span is not None:
            self.z_span = z_span
            self.vna.set_span(z_span)
        else:
            self.z_span = self.vna.get_span()
        x_fit = np.array(curve_p[0])
        z_fit = np.array(curve_p[1])
        try:
            if curve_f == 'lin_spline':
                self.xzlandscape_func = interp1d(x_fit, z_fit)
            elif curve_f == 'spline':
                self.xzlandscape_func = UnivariateSpline(x_fit, z_fit)
            else:  # curve_fit procedure differs from splines
                if curve_f == 'parab':
                    fit_fct = self.f_parab
                elif curve_f == 'hyp':
                    fit_fct = self.f_hyp
                elif curve_f == 'lin':
                    fit_fct = self.f_lin
                    p0 = p0[:2]
                else:
                    raise ValueError('function type not known...aborting')
                popt, pcov = curve_fit(fit_fct, x_fit, z_fit/multiplier, p0=p0)
                self.xzlandscape_func = lambda x: multiplier * fit_fct(x, *popt)

            self.xz_freqpoints = np.arange(np.min(self.xzlandscape_func(self.spec.x_vec)) - self.z_span / 2,
                                           np.max(self.xzlandscape_func(self.spec.x_vec)) + self.z_span / 2,
                                           self.vna.get_span() / self.vna.get_nop())
        except Exception as message:
            print('fit not successful:', message)

    def delete_landscape_function_xy(self, n=None):
        """
        delete single xy landscape function n (with 0 being the first one generated)
        or the complete landscape for n not specified
        """
        if n is not None:
            del self.xylandscapes[n]
        else:
            self.xylandscapes = []


    def delete_landscape_function_xz(self):
        """
        delets the xz_landscape_function incl. the xz_freqpoints
        :return: None
        """
        self.xzlandscape_func = None
        self.xz_freqpoints = None
        print('xz_landscape deleted')

    def plot_xy_landscape(self):
        """
        Plots the xy landscape(s) (for 3D scan, z-axis (vna) is not plotted
        :return:
        """
        if not qkit.module_available("matplotlib"):
            raise ImportError("matplotlib not found.")

        if self.xylandscapes:
            for i in self.xylandscapes:
                try:
                    arg = np.where((i['x_range'][0] <= self.spec.x_vec) & (self.spec.x_vec <= i['x_range'][1]))
                    x = self.spec.x_vec[arg]
                    t = i['center_points'][arg]
                    plt.plot(x, t, color='C1')
                    if i['blacklist']:
                        plt.fill_between(x, t + i['y_span'] / 2., t - i['y_span'] / 2., color='C3', alpha=0.5)
                    else:
                        plt.fill_between(x, t + i['y_span'] / 2., t - i['y_span'] / 2., color='C0', alpha=0.5)
                except Exception as e:
                    print(e)
                    print('invalid trace...skip')
            plt.axhspan(np.min(self.spec.y_vec), np.max(self.spec.y_vec), facecolor='0.5', alpha=0.5)
            plt.xlim(np.min(self.spec.x_vec), np.max(self.spec.x_vec))
            plt.show()
        else:
            print('No trace generated. Use landscape.generate_xy_function')

    def plot_xz_landscape(self):
        """
        plots the xz landscape, i.e., how your vna frequency span changes with respect to the x vector
        :return: None
        """
        if not qkit.module_available("matplotlib"):
            raise ImportError("matplotlib not found.")

        if self.xzlandscape_func:
            y_values = self.xzlandscape_func(self.spec.x_vec)
            plt.plot(self.spec.x_vec, y_values, 'C1')
            plt.fill_between(self.spec.x_vec, y_values+self.z_span/2., y_values-self.z_span/2., color='C0', alpha=0.5)
            plt.xlim((self.spec.x_vec[0], self.spec.x_vec[-1]))
            plt.ylim((self.xz_freqpoints[0], self.xz_freqpoints[-1]))
            plt.show()
        else:
            print('No xz funcion generated. Use landscape.generate_xz_function')

    def set_y_span(self, y_span, n=None):
        if not self.xylandscapes:
            self.y_span_default = y_span
        else:
            if n is not None:
                self.xylandscapes[n]['y_span'] = y_span
            else:
                for e in self.xylandscapes:
                    e['y_span'] = y_span

    def get_y_span(self):
        return self.y_span_default

    def get_y_span_list(self):
        return [e['y_span'] for e in self.xylandscapes]

    def vna_frequency_wrapper(self, x_set_obj):
        """
        Function wraps the x_set_obj function for a landscape_xy scan such that the vna frequencies are automatically
        adjusted
        :param x_set_obj: as given by the user
        :return: the wrapped function
        """
        def vna_wrapper(x):
            start_freq = self.xz_freqpoints[np.argmin(np.abs(self.xz_freqpoints -
                                                             (self.xzlandscape_func(x)-self.z_span/2)))]
            self.vna.set_startfreq(start_freq)
            self.vna.set_stopfreq(start_freq + self.z_span)
            x_set_obj(x)
        return vna_wrapper

    def perform_measurement_at_point(self, x, y, ix):
        """
        Looks if the point is of interest and returns True or False
        :param x: point in the x-vector
        :param y: point in the y-vector
        :param ix: index of x-vector
        :return: measure_bool, whether or not the point should be measured
        """
        # looks strange but works
        measure = False
        for e in self.xylandscapes:
            if np.abs(e['center_points'][ix] - y) <= e['y_span'] / 2 \
            and e['x_range'][0] <= x <= e['x_range'][1]:  # The point is covered by this span
                if e['blacklist']:
                    return False  # if the point is blacklisted anywhere, we don't need to look further
                else:
                    measure = True
        return measure

    def get_tracedata_xz(self, x):
        """
        Function to fill the tracedata with NaNs outside the scan region
        :param x: x_value where we perform the scan
        :return:
        """
        amp = np.full_like(self.xz_freqpoints, np.NaN, dtype=np.float16)
        pha = np.full_like(self.xz_freqpoints, np.NaN, dtype=np.float16)
        startarg = np.argmin(np.abs(self.xz_freqpoints-(self.xzlandscape_func(x)-self.z_span/2)))
        stoparg = startarg+ self.vna.get_nop()
        a, p = self.vna.get_tracedata()
        amp[startarg:stoparg] = a
        pha[startarg:stoparg] = p
        return amp, pha

    def get_freqpoints_xz(self):
        return self.xz_freqpoints

    def f_parab(self, x, a, b, c):
        return a * (x - b) ** 2 + c

    def f_hyp(self, x, a, b, c):
        "hyperbolic function with the form y = sqrt[ a*(x-b)**2 + c ]"
        return np.sqrt(a * (x - b) ** 2 + c)

    def f_lin(self, x, a, b):
        return a * x + b

    def f_transmon(self, x, w_max, L, I_ext, djj):
        """
        Dispersion of a tunable transmon qubit with junction asymmetry.

        Args:
            x:     x-value
            w_max:    Maximum qubit frequency without detuning.
            L:     Oscillation period in current/x-value.
            I_ext: Offset current/x offset.
            djj:   Josephson junction asymmetry (Ic1 - Ic2)/(Ic1 + Ic2)

        Returns:
            Primal transition frequency of a transmon qubit.
        """
        return w_max * (np.abs(np.cos(np.pi / L * (x - I_ext))) * (
                1 + djj ** 2 * np.tan(np.pi / L * (x - I_ext)) ** 2) ** .5) ** 0.5



#-------------------------------------------------------------------------------
# Name:        SendMail
# Purpose:     Sends an eMail with topic, a message and an png-File as
#              attachment to several recipients.
#
# Author:      A.Tkalcec    (andrej.tkalcec@gmail.com)
# modified:    J. Braumueller (20.06.2014)
#
# Created:     08.02.2013
#-------------------------------------------------------------------------------

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import smtplib
import time,os

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.image import MIMEImage

class SendMail(object):

    def __init__(self, source, subject):
        # Source = Name of the transmitter ["string"] shoud be in eMail
        # format like: xy@email.com (eMail does not have to exist)
        self.Source = source
        # Subject = Topic, information of the message, needs to be an array: ["string"]
        self.Subject = subject
        self.Recipients = "[jochen.braumueller@kit.edu]"
        self.Msg = MIMEMultipart()
        # For "single-line" messages
        self.lineList = []
        # To attach several images
        self.imgList = []
        #path to recipients.txt
        self.rptsPath = os.getcwd() + "/recipients.txt"
        
        
    def setRecipientsDir(self,rptsPath = os.getcwd() + "/recipients.txt"):
        self.rptsPath = rptsPath

    def setRecipients(self, rpts = ""):
 
        strTime = time.strftime('%X %x')
        if rpts != "":   #recipient specified
            try:
                self.Recipients = [rpts]
            except:
                print strTime + " sendMail: invalid recipient"
        else:   #use recipients.txt
            try:
                f1 = open(self.rptsPath, 'r')
                rpts_lines = f1.readlines()
                f1.close()
                self.Recipients = [l.strip() for l in rpts_lines]   #strip off \n
            except:
                print strTime + " sendMail: Failed to open or read recipients.txt"
            
    def add_txt_file(self, mesgPath):
        
        # mesgPath = Path of a TXT-File with the incl. information
        # Opens a plain text file for reading (ASCII)
        strTime = time.strftime('%X %x')
        try:
            f2 = open(mesgPath, 'rb')
            # Create a text/plain message
            self.lineList.append(f2.read())
            f2.close()
            # Create the body of the message.
        except IOError:
            print strTime+" sendMail: Faild to open or read txt-file for message!"
            self.add_Line("Faild to open or read txt-file for message!")


    def setImage(self, imgPath):
        
        strTime = time.strftime('%X %x')
        try:
            # imgPath = Path of a image-File as PNG ["string"]
            self.imgList.append(imgPath)
        except IOError:
            print strTime+" sendMail: Faild to read path for PNG!"
            self.add_Line("Faild to read path for PNG!")


    def sendEMail(self):
        
        strTime = time.strftime('%X %x')
        
        COMMASPACE = ', '
        # Create message container
        self.Msg['Subject'] = self.Subject
        self.Msg['From'] = self.Source
        self.Msg['To'] = COMMASPACE.join(self.Recipients)

       # Messages to concatenate
        if self.lineList != []:
             txt = MIMEText(''.join(self.lineList))
             self.Msg.attach(txt)
        else:
            txt = MIMEText("NO MESSAGE")
            self.Msg.attach(txt)

        # Attach image
        if self.imgList != []:
            try:
                for line in self.imgList:
                    f3 = open(line, 'rb')
                    img = MIMEImage(f3.read())
                    f3.close()
                    self.Msg.attach(img)
            except IOError:
                print strTime + " sendMail: Faild to open or read path for PNG!"

        # Send the message via local SMTP server.
        s = smtplib.SMTP('smtp.kit.edu',25)   #kit's smtp server, port 25
        #print "was here"
        # sendmail function takes 3 arguments: sender's address, recipient's
        # address and message to send
        s.sendmail(self.Source, self.Recipients, self.Msg.as_string())
        s.quit()
        print strTime + " sendMail: eMail sent to: " + str(self.Recipients)


    def add_Line(self, mesgLine):
        
        # Call this function for several "single-line" messages
        strTime = time.strftime('%X %x')
        try:
            #self.lineList.append("\n")
            self.lineList.append(mesgLine + '\n')
            #self.lineList.append("\n")
        except IOError:
            print strTime + " sendMail: Failed to write message to array (single-line)!"
            txt = MIMEText("Failed to write message to array (single-line)")
            self.Msg.attach(txt)

    def __del__(self):
        return True



import zmq

from qkit.core.instrument_base import Instrument


class RT_MW_Switch(Instrument):
    """
    This is the remote switch client to connect to the room temperature switch.
    Initialize with:
    <name> = instruments.create('<name>', 'RT_MW_Switch', url = "tcp://rt-switch-pi.local:5000")
    """

    def __init__(self, name, url):
        Instrument.__init__(self, name, tags=['physical'])

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.connect(url)

        self.add_function('switch_to_position')

    def switch_to_position(self, position):
        """
        Switches to a position of the microwave switch.
        In this case, position may be either 1 or 2.
        """
        assert position in (1, 2), f"Invalid position {position}!"
        self.socket.send_string(f"S{position}")
        return self.socket.recv_string()



# qtlab driver/wrapper for EdwardsActiveDigitalController by JB@KIT 03/2015
# pressure readout device @ UFO: G1, G2
# model: D39591000 (enhanced)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


from qkit.core.instrument_base import Instrument
import types
import logging
#import numpy as np
import time,sys
#import atexit
#import serial
import socket
import sys

class EdwardsActiveDigitalController(Instrument):

    def __init__(self, name, host='ip-address', port=9955):

        Instrument.__init__(self, name, tags=['physical'])
        
        self.add_parameter('condenser_pressure',
            type=types.FloatType,
            flags=Instrument.FLAG_GET, units='mbar')
        self.add_parameter('still_pressure',
            type=types.FloatType,
            flags=Instrument.FLAG_GET, units='mbar')
            
        self.HOST, self.PORT = host, port

    def do_get_condenser_pressure(self):
        try:
            # Create a socket (SOCK_STREAM means a TCP socket)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            #connect
            sock.connect((self.HOST, self.PORT))
            sock.sendall('get_p_cond' + '\n')
            # Receive data from the server and shut down
            rec = sock.recv(1024)
            sock.close()
            return rec
        except Exception as detail:
            print "Error: ", detail
            return 0
            
    def do_get_still_pressure(self):
        try:
            # Create a socket (SOCK_STREAM means a TCP socket)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            #connect
            sock.connect((self.HOST, self.PORT))
            sock.sendall('get_p_still' + '\n')
            # Receive data from the server and shut down
            rec = sock.recv(1024)
            sock.close()
            return rec
        except Exception as detail:
            print "Error: ", detail
            return 0
    
    def get_all(self):
        self.get('condenser_pressure')
        self.get('still_pressure')
 


#JB@KIT 01/2018
'''
qkit instrument driver for DC DAC LTC2666 (client)
 +- 5V dc digital to analog converter, 8 (16) channels, 16 bit amplitude resolution
'''

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


from qkit.core.instrument_base import Instrument
import types
import logging
import zerorpc

class DC_DAC_LTC2666(Instrument):
    '''
    Instrument class for DC DAC LTC2666.
    '''
    def __init__(self, name, host='ip-address', port=9931, nchannels = 16):
        '''
        Initialize qkit parameters and connect to the zerorpc server as a client.
        '''
        Instrument.__init__(self, name, tags=['physical'])
        
        self.channels = nchannels
        self.add_parameter('voltage', type=float, flags=Instrument.FLAG_GETSET, 
                units='V', channels=(0, self.channels-1), channel_prefix='ch%d_',minval=-5.,maxval=5)
        self.voltages = [0]*self.channels
        
        self.HOST, self.PORT = host, port
        self.init_connection()
    
    def init_connection(self):
        '''
        initiate connection to zerorpc server
        '''
        logging.debug(__name__ + ': initiating connection to DAC server {:s} on port {:d}'.format(self.HOST, self.PORT))
        self.c = zerorpc.Client()
        self.c.connect("tcp://{:s}:{:d}".format(self.HOST, self.PORT))

    def do_set_voltage(self, voltage, channel):
        '''
        Set dac voltage in range -5V..5V.
        '''
        if voltage>5-10./2**16:
            voltage = 5-10./2**16 #we can not set exactly +5V, but only one bit less
        for i in range(10):
          try:
              self.c.set_voltage(channel, voltage)
              self.voltages[channel]=voltage
              break
          except Exception as detail:
              if i==9:
                  logging.error('{:s}'.format(detail))
                  
    def do_get_voltage(self, channel):
        '''
        Get the current voltage setting for channel.
        '''
        return self.voltages[channel]
            
    def close_connection(self):
        logging.debug(__name__ + ': closing connection to DAC server {:s} on port {:d}'.format(self.HOST, self.PORT))
        self.c.close()



# MP/AS @ KIT 05/2017
# JSON en-/decoder for non-JSON standard data-types

import json
import types
import numpy as np
import logging
import qkit
try:
    if qkit.module_available('uncertainties'):
        import uncertainties
        uncertainties_enable = True
    else:
        uncertainties_enable = False
except AttributeError:
    try:
        import uncertainties
        uncertainties_enable = True
    except ImportError:
        uncertainties_enable = False

class QkitJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if type(obj) == np.ndarray:
            return {'dtype': type(obj).__name__, 'content': obj.tolist()}
        # if type(obj) == types.InstanceType:  # no valid synatx in python 3 and probably not needed (MMW)
        #    return {'dtype' : type(obj).__name__, 'content': str(obj.get_name())}
        if uncertainties_enable or qkit.module_available('uncertainties'):
            if type(obj) in (uncertainties.core.Variable, uncertainties.core.AffineScalarFunc):
                return {'dtype': 'ufloat',
                        'content': {'nominal_value': obj.nominal_value,
                                    'std_dev': obj.std_dev}}
        try:
            return obj._json()
        except AttributeError:
            return {'dtype': type(obj).__name__, 'content': json.JSONEncoder.default(self, obj)}


class QkitJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if 'content' in obj and 'dtype' in obj and len(obj) == 2:
            if obj['dtype'] == 'ndarray':
                return np.array(obj['content'], dtype=object)
            if obj['dtype'] == 'ufloat':
                if uncertainties_enable or qkit.module_available('uncertainties'):
                    return uncertainties.ufloat(nominal_value=obj['content']['nominal_value'],
                                                std_dev=obj['content']['std_dev'])
                else:
                    logging.warning('Uncertainties package not installed. Only nominal value returned.')
                    return float(obj['content']['nominal_value'])
            if obj['dtype'] == 'qkitInstrument':  # or obj['dtype'] == 'instance'
                return qkit.instruments.get(obj['content'])
        else:
            return obj 


# Windfreaktech SynthHD Microwave Source
# Andre Schneider <erdna.na@gmail.com>, 2015
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
import math
from time import sleep
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from distutils.version import LooseVersion


class Windfreaktech_SynthHD(Instrument):
    '''
    This is the python QTLab driver for the Windfreaktech SynthHD microwave source
    '''

    def __init__(self, name, address, model='Windfreaktech'):
        '''
        Initializes the Windfreaktech_SynthHD, and communicates with the wrapper.
        
        Input:
            name (string)    : name of the instrument
            address (string) : address
        '''
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._model = model
        self._visainstrument = visa.instrument(self._address)
        if LooseVersion(visa.__version__) < LooseVersion("1.5.0"):
            self._visainstrument.term_chars = '\n'
        else:
            self._visainstrument.write_termination = '\n'
            self._visainstrument.read_termination = '\n'
        self._numchannels = 2
        self._frequency = [None, None, None]
        self._power = [None, None, None]

        # Implement parameters
        self.add_parameter('frequency', type=float,
                           flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
                           channels=(1, self._numchannels), minval=0, maxval=13.6e9, units='Hz', channel_prefix='ch%d_')

        self.add_parameter('power_level', type=float,
                           flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
                           channels=(1, self._numchannels), minval=0, maxval=45000, units='arbu',
                           channel_prefix='ch%d_')

        self.add_parameter('power', type=float,
                           flags=Instrument.FLAG_GETSET,
                           channels=(1, self._numchannels), minval=-70, maxval=20, units='dBm', channel_prefix='ch%d_')

        self.add_parameter('status', type=bool,
                           flags=Instrument.FLAG_GETSET,
                           channels=(1, self._numchannels), channel_prefix='ch%d_')

        self.add_function('get_all')
        self.add_function('reload_calibration')
        # self.add_function('set_PLL_frequency')
        # self.add_function('get_PLL_frequency')

        self.reload_calibration()
        self.get_all()
        self.set_PLL_frequency(10e6)

    # initialization related
    def get_all(self):
        self.get_ch1_frequency()
        self.get_ch2_frequency()
        self.get_ch1_status()
        self.get_ch2_status()
        self._visainstrument.write("C0h1C1h1")
        pass

    # Communication with device

    def reload_calibration(self):
        '''
        reloads power calibration from file Windfreaktech_SynthHD.cal in instruments folder.
        '''
        try:
            data = np.loadtxt(qkit.cfg.get('instruments_dir') + '/Windfreaktech_SynthHD.cal')
            f = data[1:, 0]
            p = data[0, 1:]
            values = data[1:, 1:]
            self._interp_amplitude = RegularGridInterpolator((f, p), values).__call__

        except IOError:
            raise IOError('Calibration file for WFT MW Source not found')

    def delete(self):
        '''
        Use this to close the session with the device. It is more likely to work again afterwards.
        '''
        self.__del__()

    def __del__(self):
        self._visainstrument.close()
        print("Session closed.")

    def set_PLL_frequency(self, frequency):
        '''
        set the reference frequency for the PLL in Hz
        and set 'x' to 0 (external reference clock)
        if frequency == False: Set clock to internal reference
        '''
        if frequency == False:
            self._visainstrument.write("*%fx1" % (27))
        else:
            self._visainstrument.write("*%fx0" % (frequency / 1e6))

    def get_PLL_frequency(self):
        return float(self._visainstrument.ask("*?")) * 1e6

    def do_get_frequency(self, channel):
        self._frequency[channel] = float(self._visainstrument.ask('C%if?' % (channel - 1))) * 1e6
        return self._frequency[channel]

    def do_set_frequency(self, frequency, channel):
        '''
        Set frequency of device

        Input:
            freq (float) : Frequency in Hz

        Output:
            None
        '''
        self._frequency[channel] = frequency
        self._visainstrument.write('C%if%.9f' % (channel - 1, frequency / 1e6))
        self.do_set_power(self._power[channel], channel)

    def do_get_power_level(self, channel):
        return float(self._visainstrument.ask('C%ia?' % (channel - 1)))

    def do_set_power_level(self, level, channel):
        '''
        Set uncalibrated power level of device
        Usually, you should use set_power which accepts values in dBm

        Input:
            level (float) : (0=mimimum, 45000=maximum)

        Output:
            None
        '''
        self._visainstrument.write('C%ia%.3f' % (channel - 1, level))

    def do_get_power(self, channel):
        '''
        this only returns the stored power value and does not communicate with the device!
        '''
        if self._power[channel] == None:
            print("Power was not set since reload of driver. You have to do this to get a proper value.")
        return self._power[channel]

    def do_set_power(self, power, channel):
        '''
        Set output power of device

        Input:
            power (float) : Power in dBm

        Output:
            None
        '''
        if (np.isnan(float(self._interp_amplitude((self._frequency[channel], power))))):
            raise ValueError(
                'The possible minimum and maximum output power of this microwave source does depend on the frequency. Unfortunately, your desired combination of frequency and output power is not possible.')
        else:
            self._power[channel] = power
            self.do_set_power_level(int(self._interp_amplitude((self._frequency[channel], power))), channel)

    def do_get_status(self, channel):
        return bool(int(self._visainstrument.ask('C%ir?' % (channel - 1))))

    def do_set_status(self, status, channel):
        '''
        Set status of output channel
        Sets 'r' (output-parameter)  as well as 'E' (PLL Chip Enable) according to the manual

        Input:
            status : True or False 

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting status to "%s"' % status)
        if status == True:
            self._visainstrument.write('C%ir1E1' % (channel - 1))
        elif status == False:
            self._visainstrument.write('C%ir0E0' % (channel - 1))
        else:
            raise ValueError("set_status(): Please use either True or False")

    # shortcuts
    def off1(self):
        '''
        Set ch1 status to 'off'

        Input:
            None

        Output:
            None
        '''
        self.set_ch1_status(False)

    def off2(self):
        self.set_ch2_status(False)

    def off(self):
        '''
        turns both outputs off
        '''
        self.set_ch1_status(False)
        self.set_ch2_status(False)

    def on1(self):
        '''
        Set status to 'on'

        Input:
            None

        Output:
            None
        '''
        self.set_ch1_status(True)

    def on2(self):
        self.set_ch2_status(True)

    def write(self, msg):
        return self._visainstrument.write(msg)

    def ask(self, msg):
        return self._visainstrument.ask(msg)



# -*- coding: utf-8 -*-

# Teslameter_FM302.py driver for the Projekt Elektronic GmbH Teslameter FM302
# Micha Wildermuth, micha.wildermuth@kit.edu 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
from distutils.version import LooseVersion
import time


class Teslameter_FM302(Instrument):
    """
    This is the driver for the Projekt Elektronic GmbH Teslameter FM302
    """

    def __init__(self, name, address, reset=False):
        """
        Initializes VISA communication with the instrument Yokogawa GS820.

        Parameters
        ----------
        name: string
            Name of the instrument (driver).
        address: string
            Serial address for the communication with the instrument.
        reset: bool, optional
            Resets the instrument to default conditions. Default is False.

        Returns
        -------
        None

        Examples
        --------
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]

        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py

        >>> TM = qkit.instruments.create('TM', 'Teslameter_FM302', address='ASRL1::INSTR')
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        """
        self.__name__ = __name__
        # Start VISA communication
        Instrument.__init__(self, name, tags=['physical'])
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        # Set termination characters (necessary for Ethernet communication)
        if LooseVersion(visa.__version__) < LooseVersion("1.5.0"):  # pyvisa 1.4
            self._visainstrument.term_chars = '\n'
        else:  # pyvisa 1.5
            self._visainstrument.write_termination = '\n'
            self._visainstrument.read_termination = '\r\n'
        if reset:
            self.reset()
        self.add_parameter('coupling',
                           type=int,
                           flags=Instrument.FLAG_GETSET,
                           minval=0,
                           maxval=1)
        self.add_parameter('filter',
                           type=int,
                           flags=Instrument.FLAG_GETSET,
                           minval=1,
                           maxval=64)
        self.add_parameter('integration_time',
                           type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=100e-3,
                           maxval=5,
                           unit='s')
        self.add_parameter('gain',
                           type=int,
                           flags=Instrument.FLAG_GETSET,
                           minval=1,
                           maxval=100)
        self.add_parameter('relative',
                           type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=-70,
                           maxval=70,
                           units='T')
        self.add_parameter('zero',
                           type=int,
                           flags=Instrument.FLAG_GETSET,
                           minval=-39320,
                           maxval=26213)
        self.add_parameter('value',
                           type=float,
                           flags=Instrument.FLAG_GET,
                           minval=-1.,
                           maxval=1.,
                           units='T')
        self.add_function('reset')

    def _write(self, cmd):
        return self._visainstrument.write(cmd)

    def _read(self):
        return self._visainstrument.read_raw().decode('latin1').strip('\r\n')

    def do_set_coupling(self, val):
        """
        Sets measurement signal coupling to <val>.

        Parameters
        ----------
        val: int
            Signal coupling mode. Must be 0 (DC) or 1 (AC).

        Returns
        -------
        None
        """
        if val==0:  # DC
            self._visainstrument.query('coupling DC')
        elif val==1:  # AC
            self._visainstrument.query('coupling AC')

    def do_get_coupling(self):
        """
        Gets the measurement signal coupling <val>.

        Parameters
        ----------
        None

        Returns
        -------
        val: int
            Signal coupling mode. Meanings are 0 (DC) and 1 (AC).
        """
        return {'DC':0, 'AC':1}[self._visainstrument.query('coupling').strip('coupling is ')]

    def do_set_filter(self, val):
        """
        Sets the moving average filter length to <val>.

        Parameters
        ----------
        val: int
            Moving average filter length. Can be 1, 2, 4, 8, 16, 32, 64.

        Returns
        -------
        None
        """
        self._write('filter {:d}'.format(val))
        return int(self._read().strip('filter is '))

    def do_get_filter(self):
        """
        Gets the moving average filter length <val>.

        Parameters
        ----------
        None

        Returns
        -------
        val: int
            Moving average filter length. Possible values are 1, 2, 4, 8, 16, 32, 64.
        """
        self._write('filter')
        return int(self._read().strip('filter is '))

    def do_set_integration_time(self, val):
        """
        Sets the integration time to <val>. The internal sampling rate is 10Hz.

        Parameters
        ----------
        val: int
            Integration time in seconds. Must be in [0.1s, 25.5s].

        Returns
        -------
        None
        """
        self._write('inttime {:d}'.format(val*1e3))
        return float(self._read().strip('integration time is '))*1e-3

    def do_get_integration_time(self):
        """
        Gets the integration time <val>. The internal sampling rate is 10Hz.

        Parameters
        ----------
        None

        Returns
        -------
        val: int
            Integration time in seconds.
        """
        self._write('inttime')
        return float(self._read().strip('integration time is '))*1e-3

    def do_set_gain(self, val):
        """
        Sets the analog gain to <val>.

        Parameters
        ----------
        val: int
            Analog gain. Must be in {1, 10, 100}.

        Returns
        -------
        None
        """
        self._write('gain {:d}'.format(val))
        return int(self._read().strip('analog gain is x'))

    def do_get_gain(self):
        """
        Gets the analog gain <val>.

        Parameters
        ----------
        None

        Returns
        -------
        val: int
            Analog gain. Can be in {1, 10, 100}.
        """
        self._write('gain')
        return int(self._read().strip('analog gain is x'))

    def do_set_relative(self, val):
        """
        Sets measurement relative value.

        Parameters
        ----------
        val: float:
            Absolute measurement (0) or relative measurement (<val> or 1 for current value as relative measurement value).

        Returns
        -------
        None
        """
        if bool(val) is False:
            self._write('absolute')
        else:
            if bool(val) == val:
                self._write('relative set')
            else:
                self._write('relative {:1.2f}'.format(val))
        return self._read()

    def do_get_relative(self):
        """
        Gets measurement relative value.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Relative measurement value.
        """
        self._write('relative')
        ans = self._read()
        if ans is 'display is not relative':
            return 0
        elif 'display is relative' in ans:
            ans = ans.strip('display is relative, reference = ').split(' ')
            return float(ans[0]) * {'m': 1e-3, u'\xb5': 1e-6, 'n':1e-9}[ans[1][0]]

    def do_set_zero(self, val):
        """
        Sets the offset compensation.

        Parameters
        ----------
        val: float
            Offset compensation value.
        val: float:
            Absolute measurement (0) or relative measurement (<val> or 1 for current value as relative measurement value).

        Returns
        -------
        None
        """
        if bool(val) == val:
            self._write('zero set')
            time.sleep(5)
        else:
            self._write('zero {:d}'.format(val))
        while True:
            ans = self._read()
            if 'zero compensation value is ' in ans:
                return int(ans.strip('zero compensation value is '))
            else:
                time.sleep(1)

    def do_get_zero(self):
        """
        Gets the offset compensation.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Offset compensation value.
        """
        self._write('zero')
        return int(self._read().strip('zero compensation value is '))

    def reset(self):
        """
        Resets the instrument to factory settings.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self._write('default')
        return self._read()

    def do_get_value(self):
        """
        Gets the measurement value.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Measured value (in Tesla).
        """
        self._visainstrument.query('logging 1')
        ans = self._read().split(' ')
        return float(ans[0]) * {'m': 1e-3, u'\xb5': 1e-6, 'n':1e-9}[ans[1][0]]
